#%%
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
import glob, os, importlib, sys
os.chdir(os.path.dirname(__file__))
os.chdir("..")  # ensures we are in the main directory
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
PROJECT_ROOT = os.path.dirname(SCRIPT_DIR)
sys.path.append(PROJECT_ROOT)
from src import ACID_code as acid
skips = 1 # Skip some values to save time in this tutorial

# Quickstart Example
spec_file = fits.open('example/sample_spec_1.fits')

wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
spectrum = spec_file[1].data[::skips]     # Spectral Flux
error = spec_file[2].data[::skips]        # Spectral Flux Errors
sn = spec_file[3].data                    # SN of Spectrum

linelist_path = 'example/example_linelist.txt' # Insert path to line list

# Choose a velocity grid for the final profile(s), you can use the calc_deltav
# function to get a velocity pixel size if desired, otherwise, set your own deltav value
# the velocity pixel size must not be smaller than the spectral pixel size
velocities = np.arange(-25, 25, 0.82)

# Initiate Acid
Acid = acid.Acid(velocities=velocities, linelist_path=linelist_path)
# Run ACID
result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000)

# Plot your final profile
result.plot_profiles() # See documentation for more plot kwarg options

# You can also plot walkers, corner plots, etc. See documentation.

# If you feel the need to continue sampling, you can do so with:
# result.continue_sampling(nsteps=2000) # And plot walkers to see the difference!

#%% Multiple Frames Example

# finds sample files in 'example directory'. Each file is a different frame.
files = glob.glob('example/sample_spec_*.fits')

# create lists for wavelengths, spectra, errors and sn for all frames
wavelengths = []
spectra = []
errors = []
sns = []

for file in files:
    spec_file = fits.open('%s'%file)

    wavelengths.append(spec_file[0].data[::skips])    # Wavelengths in Angstroms
    spectra.append(spec_file[1].data[::skips])        # Spectral Flux
    errors.append(spec_file[2].data[::skips])         # Spectral Flux Errors
    sns.append(float(spec_file[3].data))     # SN of Spectrum

linelist_path = 'example/example_linelist.txt' # Insert path to line list

# choose a velocity grid for the final profile(s)
velocities = np.arange(-25, 25, 0.82)

# run ACID function
Acid = acid.Acid(velocities=velocities, linelist_path=linelist_path)
result = Acid.ACID(wavelengths, spectra, errors, sns, nsteps=2000)

# plot results
result.plot_profiles()

# Remember you can obtain the entire result array via:
all_frames = result[:] # Which converts it to a numpy array of shape (frames, orders, 2, pixels)
# or you can do
all_frames = result.all_frames

#%% Multiple Orders Example

spec_file = fits.open('example/sample_spec_1.fits')

wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
spectrum = spec_file[1].data[::skips]     # Spectral Flux
error = spec_file[2].data[::skips]        # Spectral Flux Errors
sn = spec_file[3].data                    # SN of Spectrum

linelist_path = 'example/example_linelist.txt' # Insert path to line list

# choose a velocity grid for the final profile(s)
velocities = np.arange(-25, 25, 0.82)

# choose size of wavelength ranges (or chunks)
wave_chunk = 25
chunks_no = int(np.floor((max(wavelength)-min(wavelength))/wave_chunk))
min_wave = min(wavelength)
max_wave = min_wave+wave_chunk

# create result array of shape (no. of frames, no. of chunks, 2, no. of velocity pixels)
result = np.zeros((1, chunks_no, 2, len(velocities)))

# Initiate Acid
Acid = acid.Acid(velocities=velocities, linelist_path=linelist_path)

for i in range(chunks_no):

    # use indexing to select correct chunk of spectrum
    idx = np.logical_and(wavelength>=min_wave, wavelength<=max_wave)

    # You can recursively call ACID to fill in each order
    # In the future, this loop will be handled internally by ACID and this
    # example will be updated accordingly.
    result = Acid.ACID(wavelength[idx], spectrum[idx], error[idx], sn,
                           all_frames=result, order=i, nsteps=2000)

    min_wave += wave_chunk
    max_wave += wave_chunk

# plot results
result.plot_profiles()
# Reember you can save and load results via:
# result.save_result("example/multi_order_result.pkl")
# result = acid.Result.load_result("example/multi_order_result.pkl")

#%% HARPS data example

e2ds_files = glob.glob('tests/data/*e2ds_A*.fits') # Returns list of HARPS files
linelist_path = 'example/example_linelist.txt'
save_path = 'no save'
order_range = np.arange(41, 43) # Specify which orders to run ACID on (here we do 41 and 42 as an example)

# choose a velocity grid for the final profile(s)
deltav = 0.82     # velocity pixel size for HARPS e2ds data from DRS pipeline 3.5
velocities = np.arange(-25, 25, deltav)

# run ACID function
Acid = acid.Acid(velocities=velocities, linelist_path=linelist_path)

# Due to legacy behaviour, the function returns BJDs, profiles and errors separately when indexed,
# not all_frames as in other examples. All frames can still be accessed via result.all_frames

result = Acid.ACID_HARPS(filelist=e2ds_files, file_type='e2ds', save_path=save_path, nsteps=2000,
                             order_range=order_range)

# BJDs, profiles, profile_errors = result
all_frames = result.all_frames
result.plot_profiles()



from __future__ import annotations
import numpy as np
from . import utils
from beartype import beartype
from numpy.polynomial import polynomial as P
from scipy.linalg import cho_solve

# The following two wrapper functions are required for multiprocessing
# support, without it, the fork method would need to reserialize everything
# which is very inefficient. See parallelization in the emcee documentation
# for more details.
def _mp_init_worker(data):
    """Initializes each worker process with global data."""
    global _MCMC
    _MCMC = MCMC(data)

def _mp_log_probability(theta):
    """Wrapper for log probability function for multiprocessing."""
    return _MCMC(theta)

class MCMC:

    @beartype
    def __init__(
            self,
            x_or_data   : np.ndarray|object,
            y           : np.ndarray|None = None,
            yerr        : np.ndarray|None = None,
            alpha       : np.ndarray|None = None,
            velocities  : np.ndarray|None = None,
            c_factor                      = None,
            deterministic_profile : bool  = False,
        ):
        """Initialise MCMC functions with necessary data.
        Called once per worker if using multiprocessing.

        Parameters
        ----------
        x_or_data : np.ndarray|object
            Wavelength array or Data object. If a Data object is provided, takes all 
            the arguments below from there. If a Data object is provided, all other 
            arguments are ignored.
        y : np.ndarray, optional
            Observed flux array, required if x_or_data is a np.ndarray.
        yerr : np.ndarray, optional
            Observed flux error array, required if x_or_data is a np.ndarray.
        alpha : np.ndarray, optional
            Precomputed alpha matrix, required if x_or_data is a np.ndarray.
        velocities : np.ndarray|None, optional
            Velocity grid for LSD profile, only needed when calling log-probability
            function, by default None.
        c_factor : optional
            Precomputed c_factor for LSD profile calculation, by default None.
        deterministic_profile : bool, optional
            Whether to fit the full profile (True) or use the fast model (False), by default True.
        """

        # No checks are performed here - assume data is valid from ACID class checks,
        # else user is on their own!
        if not isinstance(x_or_data, np.ndarray):
            data = x_or_data
            self.x = data.wavelengths["masked"]
            self.y = data.flux["masked"]
            self.yerr = data.errors["masked"]
            self.alpha = data.alpha
            self.velocities = data.velocities
            self.c_factor = data.c_factor
            self.deterministic_profile = data.config.deterministic_profile

        else:
            self.x = x_or_data
            self.y = y
            self.yerr = yerr
            self.alpha = alpha
            self.velocities = velocities
            self.c_factor = c_factor
            self.deterministic_profile = deterministic_profile

        self.k_max = self.alpha.shape[1] # the number of velocity points in the profile

        # Precompute normalization coefficients these are used to adjust the wavelengths to
        # between -1 and 1 - makes the continuum coefficents smaller and easier for emcee to handle.
        a, b = utils.get_normalisation_coeffs(self.x)
        self.u = (a * self.x) + b

        # For deterministic model, the below variables are used, and are precomputed for speed
        err_od = self.yerr / self.y # independent of continuum, since it's a ratio
        V = 1.0 / (err_od ** 2) # variance vector in log space, error already in log space
        self.AtV = self.alpha.T * V # precompute alpha matrix multiplication for _mcmc_solve_z input

        # Configure whether to use full or deterministic model
        if self.deterministic_profile is False:
            self.model_function = self.full_model # include profile fitting
        else:
            self.model_function = self.deterministic_model # infer profile points from continuum

    def __call__(self, *args, **kwargs):
        # Sets the default call is the log_probability function
        return self.log_probability(*args, **kwargs)

    def full_model(self, theta):
        """Full model for mcmc - takes all inputs (profile points + continuum coefficents)
        to create a model spectrum.

        Parameters
        ----------
        theta : array-like
            Model parameters: first k_max values are profile points (z),
            followed by continuum polynomial coefficients and scale factor.
        
        Returns
        -------
        tuple
            Model spectrum and profile points (z).
        """
        # Extract profile points and continuum coefficients from theta
        z = theta[:self.k_max]
        mdl = self.alpha @ z

        # Converting model from optical depth to flux
        mdl = np.exp(-mdl)

        # Calculate continuum polynomial
        coefs = np.asarray(theta[self.k_max:], dtype=float)

        # Apply continuum model
        mdl *= P.polyval(self.u, coefs)

        return mdl, z

    def deterministic_model(self, theta):
        """Deterministic model for mcmc - takes only continuum coefficents and scale factor
        to create a model spectrum, solving for the profile points (z) directly.

        Parameters
        ----------
        theta : array-like
            Model parameters: continuum polynomial coefficients followed by scale factor.
        
        Returns
        -------
        tuple
            Model spectrum and profile points (z).
        """
        coefs = np.asarray(theta, dtype=float)

        # Build continuum model
        mdl = P.polyval(self.u, coefs)

        if np.any(mdl <= 0): # force positive continuum at all points
            return mdl, np.full(self.k_max, -2) # return very low z to trigger prior rejection

        # Calculate fitted flux and convert to OD
        fitted_flux = self.y/mdl
        flux_od = - np.log(fitted_flux)

        # Solve for the profile points
        z = cho_solve(self.c_factor, self.AtV @ flux_od)

        # Convert back from optical depth to flux
        forward = np.exp(- (self.alpha @ z)) * mdl

        return forward, z

    def run_model_function(self, *args, **kwargs):
        """Runs the selected model function (full or fast) with given arguments.

        Returns
        -------
        tuple
            Model spectrum and profile points (z).
        """
        return self.model_function(*args, **kwargs)

    def log_prior(self, z):
        """Calculates the log prior probability of the profile points (z) and imposes the prior
        restrictions on the inputs - rejects if profile point is less than -0.4 or greater than 1.6.

        Parameters
        ----------
        z : array-like
            Profile points.

        Returns
        -------
        float
            Log prior probability.
        """

        # Hard box prior on each z[i]
        if np.any((z < -0.4) | (z > 1.6)):
            return -np.inf

        # # excluding the continuum points in the profile (in flux)
        # z_cont = []
        # v_cont = []
        # for i in range(0, 5):
        #         z_cont.append(np.exp(z[len(z)-i-1])-1)
        #         v_cont.append(self.velocities[len(self.velocities)-i-1])
        #         z_cont.append(np.exp(z[i])-1)
        #         v_cont.append(self.velocities[i])

        # z_cont = np.array(z_cont)
        # v_cont = np.array(v_cont)

        # p_pent = np.sum((np.log((1/np.sqrt(2*np.pi*0.01**2)))-0.5*(z_cont/0.01)**2))

        return 0 

    def log_probability(self, theta):
        """Calculates log probability depending on which model (full or fast).
        
        Parameters
        ----------
        theta : array-like
            Model parameters.
        
        Returns
        -------
        float
            Log probability.
        """
        forward, z = self.model_function(theta)
    
        lp = self.log_prior(z)
        if not np.isfinite(lp):
            return -np.inf

        diff = self.y - forward
        ll = -0.5 * np.sum(diff*diff / (self.yerr*self.yerr) + np.log(2*np.pi*(self.yerr*self.yerr)))
        return lp + ll

    @staticmethod
    def get_mcmc_stopping_criterion(tau_list, step_number, min_checks, min_tau_factor, tau_rel_tol):
        # --- always try to compute a tolerance metric as early as possible ---
        tol_metric = np.nan  # show NaN until computable
        tol_converged = False
        n_eff = 0

        if len(tau_list) >= 2:
            # compute a "most recent" change metric from the last two tau estimates
            a, b = tau_list[-2], tau_list[-1]
            if np.all(np.isfinite(a)) and np.all(np.isfinite(b)):
                den = np.maximum(b, 1e-12)
                tol_metric = float(np.percentile(np.abs(b - a) / den, 90))
            
            n_eff = step_number / np.max(tau_list)
            try:
                n_eff = int(n_eff)
            except:
                pass
            if np.isnan(tol_metric):
                return False, np.inf, n_eff

        # Need enough tau estimates to do min_checks consecutive deltas
        if len(tau_list) < (min_checks + 1):
            return False, np.inf, n_eff

        tau = tau_list[-1]
        if not np.all(np.isfinite(tau)):
            return False, tol_metric, n_eff
        if np.max(tau) <= 1:
            return False, tol_metric, n_eff

        tau_ref = np.max(tau)
        if step_number < 10 * tau_ref:
            return False, tol_metric, n_eff
        if not (step_number > min_tau_factor * tau_ref):
            return False, tol_metric, n_eff

        # Stability over last min_checks deltas
        recent_tau = tau_list[-(min_checks + 1):]
        rel_changes = []
        for a, b in zip(recent_tau[:-1], recent_tau[1:]):
            den = np.maximum(b, 1e-12)
            rel_changes.append(np.percentile(np.abs(b - a) / den, 90))

        tol_metric = float(np.max(rel_changes))  # overwrite with the stricter metric
        return (tol_metric < tau_rel_tol), tol_metric, n_eff

    @staticmethod
    def get_tqdm_desc(last_tolerance, last_neff, config):
        tol_str = "<" if last_tolerance < config.tau_tol else ">"
        tol_str = f"{last_tolerance:.4f}{tol_str}{config.tau_tol}"
        neff_str = ">" if last_neff > config.min_tau_factor else "<"
        neff_str = f"{last_neff:.2f}{neff_str}{config.min_tau_factor}"
        return tol_str, neff_str


from __future__ import annotations
import numpy as np
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from beartype import beartype
from scipy.special import wofz
from .utils import Array1D
from .data import Data

@beartype
class Profiles:
    """A class for fitting spectral line profiles such as Voigt and Gaussian profiles.
    """
    def __init__(
            self,
            velocities : Array1D = None,
            flux       : Array1D = None,
            flux_err   : Array1D = None,
            data       : Data    = None,
        ) -> None:
        """Initializes the Profiles class with velocity, flux, and optional flux error data.

        Parameters
        ----------
        velocities : Array1D
            The velocity values corresponding to the spectral line profile.
            Must be provided if no data instance is passed, by default None.
        flux : Array1D
            The flux values of the spectral line profile
            Must be provided if no data instance is passed, by default None.
        flux_err : Array1D, optional
            The errors associated with the flux values, by default None. If not
            input, they won't be used in the fitting process.
        data : Data, optional
            A data instance to draw velocities, flux and flux errors. Will raise an
            exception if they do not exist within the class.
            Must be provided if all three of the above inputs were not passed, by default None. 
        """

        if data is not None:
            if not getattr(data, 'velocities', None) or not getattr(data, 'profiles', None):
                raise ValueError("Data instance must have attributes 'velocities' and 'profiles'. Try running ACID first.")
            velocities = data.velocities
            flux = data.profiles[0,0] # Subtract 1 to convert from normalized flux to absorption depth
            flux_err = data.profiles[0,1]
        else:
            if velocities is None or flux is None:
                raise ValueError("If no data instance is provided, then at least velocities and flux must be provided.")

        flux -= 1 # Subtract 1 to convert from normalized flux to absorption depth

        self.velocities = velocities
        self.flux = flux
        self.flux_err = flux_err

        self.fitted_y    = {}
        self.fitted_yerr = {}
        self.fit_on_x    = {} # Store fitted values on original x for residuals

        self.fitted_x = np.linspace(np.min(velocities), np.max(velocities), 1000)
        pass

    def plot_fit(self, model:str|None='voigt', return_fig=False, **kwargs):
        """Plots the original data and the fitted profile if available.
        
        Parameters
        ----------
        model : str | None, optional
            The type of model to plot. String options are 'voigt', 'gaussian', 
            'lorentzian', 'none', or 'all'. Choosing 'none' or None will plot whichever models have 
            already been fitted for, by default 'all'.
        return_fig : bool, optional
            Whether to return the (fig, ax) tuple and not call plt.show(). If False, calls
            plt.show() and returns None. By default False.
        **kwargs : dict
            Additional keyword arguments to pass to the fitting functions if the models
            have not been fitted yet.
        Returns
        -------
        tuple | None
        """
        models = ["voigt", "gaussian", "lorentzian", "none"]
        if model is not None:
            model = model.lower()
            if model not in models + ['all']:
                raise ValueError("Model must be 'voigt', 'gaussian', 'lorentzian' or 'all'.")

        if model == 'all':
            model_list = models[:-1] # Exclude 'none'
        elif model is None or model == "none":
            model_list = list(self.fitted_y.keys())
        else:
            model_list = [model]

        for m in model_list:
            if m not in self.fitted_y:
                fit_func = getattr(self, f"fit_{m}")
                fit_func(**kwargs)

        # Plotting
        fig, ax = plt.subplots(2, 1, figsize=(8, 10), gridspec_kw={'height_ratios': [3, 1]})
        ax[0].errorbar(self.velocities, self.flux, yerr=self.flux_err, fmt='b.', label='ACID Profile', color='C0')
        ax[1].axhline(0, color='black', linestyle='--')

        for i, model in enumerate(self.fitted_y.keys()):
            y_fit = self.fitted_y[model]
            y_fit_on_x = self.fit_on_x[model]
            y_err_lo, y_err_hi = self.fitted_yerr[model]
            ax[0].plot(self.fitted_x, y_fit, label=f'{model.capitalize()} Fit', color=f'C{i+1}')
            ax[1].plot(self.velocities, y_fit_on_x - self.flux, label=f'{model.capitalize()} Residuals', color=f'C{i+1}')
        ax[1].set_xlabel('Velocity')
        ax[0].set_title('Profile Fit(s)')
        ax[1].set_ylabel('Flux')
        ax[0].set_ylabel('Flux')
        ax[0].legend()
        ax[1].legend()
        plt.show()

    def fit_voigt(self, x=None, y=None, yerr=None, p0=None, **kwargs):
        """Fits a Voigt profile to the given data.

        Parameters
        ----------
        x : array_like, optional
            The x values of the data. If None, uses self.velocities.
        y : array_like, optional
            The y values of the data. If None, uses self.flux.
        yerr : array_like, optional
            The y errors of the data. If None, uses self.flux_err.
        p0 : list, optional
            Initial guess for the parameters [amplitude, centre, sigma, gamma, offset], by default None.
        **kwargs : dict
            Additional keyword arguments to pass to curve_fit.

        Returns
        -------
        tuple
            A tuple containing the optimal parameters and the covariance matrix.
        """
        x, y, yerr = self._copy_inputs(x, y, yerr)

        if p0 is None:
            amplitude_guess = np.min(y)
            centre_guess = x[np.argmin(y)]
            sigma0 = (x.max() - x.min()) / 10.0
            gamma0 = sigma0
            offset = 0
            p0 = [amplitude_guess, centre_guess, sigma0, gamma0, offset]

        popt, pcov = self._fit_model("voigt", x, y, yerr, p0, **kwargs)
        return popt, pcov

    def fit_gaussian(self, x=None, y=None, yerr=None, p0=None, **kwargs):
        """Fits a Gaussian profile to the given data.

        Parameters
        ----------
        x : array_like, optional
            The x values of the data. If None, uses self.velocities.
        y : array_like, optional
            The y values of the data. If None, uses self.flux.
        yerr : array_like, optional
            The y errors of the data. If None, uses self.flux_err.
        p0 : list, optional
            Initial guess for the parameters [amplitude, mean, stddev], by default None.
        **kwargs : dict
            Additional keyword arguments to pass to curve_fit.

        Returns
        -------
        tuple
            A tuple containing the optimal parameters and the covariance matrix.
        """
        x, y, yerr = self._copy_inputs(x, y, yerr)

        if p0 is None:
            amplitude_guess = np.min(y)
            mean_guess = x[np.argmin(y)]
            stddev_guess = (x.max() - x.min()) / 10.0
            offset = 0
            p0 = [amplitude_guess, mean_guess, stddev_guess, offset]

        popt, pcov = self._fit_model("gaussian", x, y, yerr, p0, **kwargs)
        return popt, pcov

    def fit_lorentzian(self, x=None, y=None, yerr=None, p0=None, **kwargs):
        """Fits a Lorentzian profile to the given data.

        Parameters
        ----------
        x : array_like, optional
            The x values of the data. If None, uses self.velocities.
        y : array_like, optional
            The y values of the data. If None, uses self.flux.
        yerr : array_like, optional
            The y errors of the data. If None, uses self.flux_err.
        p0 : list, optional
            Initial guess for the parameters [amplitude, centre, gamma, offset], by default None.
        **kwargs : dict
            Additional keyword arguments to pass to curve_fit.

        Returns
        -------
        tuple
            A tuple containing the optimal parameters and the covariance matrix.
        """
        x, y, yerr = self._copy_inputs(x, y, yerr)

        if p0 is None:
            amplitude_guess = np.min(y)
            centre_guess = x[np.argmin(y)]
            gamma0 = (x.max() - x.min()) / 10.0
            offset = 0
            p0 = [amplitude_guess, centre_guess, gamma0, offset]

        popt, pcov = self._fit_model("lorentzian", x, y, yerr, p0, **kwargs)
        return popt, pcov

    def _copy_inputs(self, x, y, yerr):
        """Internal method to copy input data or use class attributes.

        Parameters
        ----------
        x : array_like | None
            The x values of the data. If None, uses self.velocities.
        y : array_like | None
            The y values of the data. If None, uses self.flux.
        yerr : array_like | None
            The y errors of the data. If None, uses self.flux_err.

        Returns
        -------
        tuple
            A tuple containing the copied x, y, and yerr arrays.
        """
        x    = np.copy(self.velocities) if x    is None else x
        y    = np.copy(self.flux)       if y    is None else y
        yerr = self.flux_err   if yerr is None else yerr
        return x, y, yerr

    def _fit_model(self, model_name, x, y, yerr, p0, **kwargs):
        """Internal method to fit a specified model to the data.

        Parameters
        ----------
        model_name : str
            The name of the model to fit ('voigt', 'gaussian', 'lorentzian').
        x : array_like
            The x values of the data.
        y : array_like
            The y values of the data.
        yerr : array_like
            The y errors of the data.
        p0 : list
            Initial guess for the parameters.
        **kwargs : dict
            Additional keyword arguments to pass to curve_fit.
        
        Returns
        -------
        tuple
            A tuple containing the optimal parameters and the covariance matrix.
        """
        # Get the model function
        model_func = getattr(self, f"{model_name}_func")

        # Perform the curve fitting
        popt, pcov = curve_fit(model_func, x, y, sigma=yerr, p0=p0, **kwargs)
        self.fitted_y[model_name] = model_func(self.fitted_x, *popt)
        self.fit_on_x[model_name] = model_func(x, *popt)

        # Get errors
        samples = np.random.multivariate_normal(mean=popt, cov=pcov, size=1000)
        y_samples = np.array([model_func(self.fitted_x, *sample) for sample in samples])
        y_lo, y_med, y_hi = np.quantile(y_samples, [0.16, 0.50, 0.84], axis=0)
        self.fitted_yerr[model_name] = (y_med - y_lo, y_hi - y_med)

        return popt, pcov

    @staticmethod
    def voigt_func(x, amplitude, centre, sigma, gamma, offset=0):
        """Calculates the Voigt profile at given x values.

        Parameters
        ----------
        x : array_like
            The x values where the Voigt profile is evaluated.
        amplitude : float
            The amplitude of the Voigt profile.
        centre : float
            The center position of the Voigt profile.
        sigma : float
            The Gaussian standard deviation.
        gamma : float
            The Lorentzian half-width at half-maximum.
        offset : float, optional
            The continuum offset, by default 0.

        Returns
        -------
        array_like
            The Voigt profile evaluated at the input x values.
        """
        z = ((x - centre) + 1j*gamma) / (sigma * np.sqrt(2))
        voigt_profile = amplitude * np.real(wofz(z)) / (sigma * np.sqrt(2*np.pi))
        return voigt_profile + offset

    @staticmethod
    def gaussian_func(x, amplitude, mean, stddev, offset=0):
        """Calculates the Gaussian profile at given x values.

        Parameters
        ----------
        x : array_like
            The x values where the Gaussian profile is evaluated.
        amplitude : float
            The amplitude of the Gaussian profile.
        mean : float
            The mean (center) of the Gaussian profile.
        stddev : float
            The standard deviation of the Gaussian profile.
        offset : float, optional
            The continuum offset, by default 0.

        Returns
        -------
        array_like
            The Gaussian profile evaluated at the input x values.
        """
        return amplitude * np.exp(-((x - mean) ** 2) / (2 * stddev ** 2)) + offset
    
    @staticmethod
    def lorentzian_func(x, amplitude, centre, gamma, offset=0):
        """Calculates the Lorentzian profile at given x values.

        Parameters
        ----------
        x : array_like
            The x values where the Lorentzian profile is evaluated.
        amplitude : float
            The amplitude of the Lorentzian profile.
        centre : float
            The center position of the Lorentzian profile.
        gamma : float
            The half-width at half-maximum.
        offset : float, optional
            The continuum offset, by default 0.

        Returns
        -------
        array_like
            The Lorentzian profile evaluated at the input x values.
        """
        return (amplitude * (gamma**2)) / ((x - centre)**2 + gamma**2) + offset


#%%
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
import glob, os, sys
from pathlib import Path
from time import time
SCRIPT_DIR = Path(__file__).resolve().parent
PROJECT_ROOT = next(p for p in SCRIPT_DIR.parents if (p / "pyproject.toml").exists())
sys.path.append(str(PROJECT_ROOT))
os.chdir(PROJECT_ROOT)
from src import  ACID_code as acid
np.random.seed(0) # Set random seed for reproducibility in tests
start = time()
skips = 5

def legacy_test():

    def quickstart():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        result = acid.ACID(wavelength, spectrum, error, linelist, sn, velocities, nsteps=2000)

        # extract profile and errors
        profile = result[0, 0]
        profile_error = result[0, 1]

        # plot results
        plt.figure()
        plt.errorbar(velocities, profile, profile_error)
        plt.xlabel('Velocities (km/s)')
        plt.ylabel('Flux')
        plt.show()
        return result

    def multiple_frames():

        # finds sample files in 'example directory'. Each file is a different frame.
        files = glob.glob('example/sample_spec_*.fits')

        # create lists for wavelengths, spectra, errors and sn for all frames
        wavelengths = []
        spectra = []
        errors = []
        sns = []

        for file in files:
            spec_file = fits.open('%s'%file)

            wavelengths.append(spec_file[0].data[::skips])    # Wavelengths in Angstroms
            spectra.append(spec_file[1].data[::skips])        # Spectral Flux
            errors.append(spec_file[2].data[::skips])         # Spectral Flux Errors
            sns.append(float(spec_file[3].data[0]))     # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelengths[0])
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        result = acid.ACID(wavelengths, spectra, errors, linelist, sns, velocities, nsteps=2000)

        # plot results
        plt.figure()

        for frame in range(len(files)):
            profile = result[frame, 0]
            profile_error = result[frame, 1]
            plt.errorbar(velocities, profile, profile_error, label = '%s'%frame)

        plt.xlabel('Velocities (km/s)')
        plt.ylabel('Flux')
        plt.legend()
        plt.show()
        return result

    _ = quickstart() # generic quick test
    _ = multiple_frames()

def class_test():
    def classes_test(skips, nsteps):
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data   # Wavelengths in Angstroms
        spectrum = spec_file[1].data     # Spectral Flux
        error = spec_file[2].data        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        velocities = np.arange(-25, 25, 0.82)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=nsteps, skips=skips)
        return result

    nsteps1 = 5000
    nsteps2 = 3000
    result = classes_test(skips=3, nsteps=nsteps1) # test the classes and result handling, use lower skips
    os.makedirs("tests/test_data", exist_ok=True) # create test data directory if it doesn't exist
    result.save_result(filename="tests/test_data/classes_test.pkl")
    result = acid.Result.load_result("tests/test_data/classes_test.pkl")
    result.plot_corner()
    result.plot_profiles()
    result.plot_walkers()
    result.plot_forward_model()
    result.continue_sampling(nsteps=nsteps2)
    assert result.sampler.get_chain().shape[0] == nsteps1 + nsteps2 and result.data.nsteps == nsteps1 + nsteps2, \
    f"Continue sampling did not add the correct number of steps to the chain.\n" \
    f"Expected {nsteps1 + nsteps2}, got {result.sampler.get_chain().shape[0]} for sampler and {result.data.nsteps} for data."  
    result.plot_walkers()
    result.plot_autocorrelation()
    result.plot_acf()
    acid.Profiles(velocities=np.arange(-25, 25, 0.82), flux=result[0,0]).plot_fit("all")
    result.plot_profiles()
    del result # clean up memory after test

def verbosity_test():
    def no_verbosity():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist, verbose=False)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000)
        return result

    def high_verbosity():
        spec_file = fits.open('example/sample_spec_1.fits')

        wavelength = spec_file[0].data[::skips]   # Wavelengths in Angstroms
        spectrum = spec_file[1].data[::skips]     # Spectral Flux
        error = spec_file[2].data[::skips]        # Spectral Flux Errors
        sn = spec_file[3].data           # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        deltav = acid.calc_deltav(wavelength)  # velocity pixel size must not be smaller than the spectral pixel size
        velocities = np.arange(-25, 25, deltav)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist, verbose=3)
        result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000)
        return result
    
    print("Starting no verbosity, until no verbosity is printed, nothing should be output")
    res_nv = no_verbosity()
    res_nv.continue_sampling(nsteps=1000)
    del res_nv # clean up memory after test
    print("End no verbosity, starting high verbosity")
    res_hv = high_verbosity()
    del res_hv # clean up memory after test

def deterministic_test():
    def deterministic_profile_fit(skips=5, nsteps=2000):
        files = glob.glob('example/sample_spec_*.fits')

        # create lists for wavelengths, spectra, errors and sn for all frames
        wavelengths = []
        spectra = []
        errors = []
        sns = []

        for file in files:
            spec_file = fits.open('%s'%file)

            wavelengths.append(spec_file[0].data)    # Wavelengths in Angstroms
            spectra.append(spec_file[1].data)        # Spectral Flux
            errors.append(spec_file[2].data)         # Spectral Flux Errors
            sns.append(float(spec_file[3].data[0]))     # SN of Spectrum

        linelist = 'example/example_linelist.txt' # Insert path to line list

        # choose a velocity grid for the final profile(s)
        velocities = np.arange(-25, 25, 0.82)

        # run ACID function
        Acid = acid.Acid(velocities=velocities, linelist=linelist)
        result = Acid.ACID(wavelengths, spectra, errors, sns, nsteps=nsteps, parallel=True, deterministic_profile=False, skips=skips)
        result.plot_walkers()
        return result
    
    res_deterministic_fit = deterministic_profile_fit(skips=3, nsteps=2000)
    res = res_deterministic_fit
    res.plot_walkers()
    res.plot_profiles()
    res.plot_corner()
    res.plot_autocorrelation()
    res.plot_acf()

def data_and_convergence_test():
    spec_file = fits.open('example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    sn = spec_file[3].data           # SN of Spectrum

    linelist = 'example/example_linelist.txt' # Insert path to line list

    # choose a velocity grid for the final profile(s)
    velocities = np.arange(-25, 25, 0.82)

    # run ACID function
    Acid = acid.Acid(velocities=velocities, linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000, skips=skips)

    data = result.data

    Acid = acid.Acid(data=data) # data does not store the sampler
    result2 = Acid.ACID(nsteps=1000, parallel=True)

    assert result2.data.nsteps == 3000, "Continue sampling did not add the correct" \
    " number of steps to the chain when using data class."

    Acid.ACID(max_steps=5000) # test new convergence function

    # test convergence check interval
    Acid = acid.Acid(data=data)
    data.config.deterministic_profile = True
    Acid.ACID(max_steps=5000)

    return

def test_edge_cases():
    spec_file = fits.open('example/sample_spec_1.fits')

    wavelength = spec_file[0].data   # Wavelengths in Angstroms
    spectrum = spec_file[1].data     # Spectral Flux
    error = spec_file[2].data        # Spectral Flux Errors
    # sn = spec_file[3].data         # SN of Spectrum

    linelist = 'example/example_linelist.txt' # Insert path to line list
    full_linelist = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9), invalid_raise=False)
    wl = full_linelist[:,0]
    depths = full_linelist[:,1]
    wl[199] = np.nan # introduce a nan value to test handling of nans in linelist
    depths[200] = np.nan # introduce a nan value to test handling of nans in linelist
    spectrum[201] = np.nan # introduce a nan value to test handling of nans in spectrum
    error[190] = np.nan # introduce a nan value to test handling of nans in error array

    # choose a velocity grid for the final profile(s)
    # velocities = np.arange(-25, 25, 0.82)

    # run ACID function
    Acid = acid.Acid(linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, nsteps=2000, 
                       skips=skips, parallel=False, deterministic_profile=True)

    data = result.data
    # continue sampling
    Acid = acid.Acid(data=data)
    result = Acid.ACID(max_steps=5000) # test continue sampling with no parallelisation

    result.continue_sampling(max_steps=5000)
    result.plot_profiles()
    result.plot_walkers()
    result.plot_autocorrelation()

    # Plot continuum fit with the data class
    result.data.plot_continuum_fit("initial")
    result.data.plot_continuum_fit("masked")
    result.data.plot_residual_masking()

    # per pixel snr, and no run_mcmc test
    sn = np.random.normal(loc=100, scale=10, size=spectrum.shape) # create a random sn array with the same shape as the spectrum
    Acid = acid.Acid(linelist=linelist)
    result = Acid.ACID(wavelength, spectrum, error, sn, nsteps=2000, run_mcmc=False,
                       skips=skips, parallel=False, deterministic_profile=True)
    assert result is None, "When run_mcmc is set to False, the ACID function should return None, but it did not."

    # Guess sn for multiple frames, plot multiple frames
    files = glob.glob('example/sample_spec_*.fits')
    wavelengths = []
    spectra = []
    errors = []
    sns = []
    for file in files:
        spec_file = fits.open('%s'%file)
        wavelengths.append(spec_file[0].data[::skips])    # Wavelengths in Angstroms
        spectra.append(spec_file[1].data[::skips])        # Spectral Flux
        errors.append(spec_file[2].data[::skips])         # Spectral Flux Errors
        sns.append(float(spec_file[3].data[0]))     # SN of Spectrum
    deltav = acid.calc_deltav(wavelengths[0])
    velocities = np.arange(-25, 25, deltav)
    result = acid.ACID(wavelengths, spectra, errors, linelist, sns, velocities, nsteps=2000)
    result.plot_profiles()

    # Test using a different moveset:
    Acid = acid.Acid(data=data)
    data.config.moves = [("StretchMove", 0.6, {}), ("DEMove", 0.2)]
    result = Acid.ACID(max_steps=5000)
    result.plot_walkers()

    # Test plotting masking lines
    result.data.config.plot_masking_lines()

    # Test plotting linelist
    result.data.plot_linelist()

    # Own telluric lines
    pass

print("Starting tests, this will take a 4-6 minutes to run, and a bunch of output will be printed.")

# The first five tests use legacy ACID inputs and calls
legacy_test()

# Now test classes
class_test()

# Test verbosities
verbosity_test()

# Test deterministic profile fit
deterministic_test()

# Add a skipping calculations using the data class test
data_and_convergence_test()

# Test edge cases, including no parallelization
test_edge_cases()


print("All tests passed!")
print(f"Total time: {time() - start:.2f} seconds")

# %%



from __future__ import annotations
from time import time
import numpy as np
import matplotlib.pyplot as plt
import corner, sys, os, pickle, warnings, contextlib, functools, inspect, psutil
from emcee import EnsembleSampler
import emcee.backends.backend as emceebackend
from beartype import beartype
from scipy.interpolate import interp1d
from numpy.polynomial import polynomial as P
from .lsd import LSD
from . import mcmc
from . import utils
from .data import Data
from .data import Config
from .utils import FloatLike, IntLike, Scalar, Array1D, Array2D, ArrayAnyD

warnings.filterwarnings("ignore")

def _require_profiles(method):
    # Make sure all results are processed before calling method
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.profiles is None:
            name = method.__qualname__
            if self.sampler is not None and self.data is not None:
                if self.config.verbose>0:
                    print(f"Note: The Result object was created without the profiles processed. " \
                        f"Running {name} requires all results to be processed, " \
                        "so process_results() will be called automatically...")
                self.process_results()
            else:
                error = f"Cannot call {name}. The profiles attribute is not available, and no " \
                "sampler and data objects are available to process results. Please pass an Acid " \
                "object after running ACID to the results init."
                raise ValueError(error)
        return method(self, *args, **kwargs)
    return wrapper

def _require_data(method):
    # Make sure Data object is available before calling method
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if self.data is None:
            name = method.__qualname__
            error = f"Cannot call {name}. The Data object is not available in this " \
            "Result instance."
            raise ValueError(error)
        return method(self, *args, **kwargs)
    return wrapper

def _require_sampler(method):
    # Make sure sampler object is available before calling method
    # TODO: Make save for a thinned chain, discard sampler unless store_sampler is True
    sig = inspect.signature(method)
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        bound = sig.bind_partial(self, *args, **kwargs)

        # A specific carvout for the save function
        store_sampler_in_args = "store_sampler" in sig.parameters
        if store_sampler_in_args is True:
            if bound.arguments.get("store_sampler", True) is False:
                return method(self, *args, **kwargs)

        sampler_in_args = "sampler" in sig.parameters
        inputted_sampler = bound.arguments.get("sampler", None)

        self.initiate_sampler(inputted_sampler if sampler_in_args else None)

        return method(self, *args, **kwargs)
    return wrapper

@beartype
class Result:
    """Class to handle the results from the Acid MCMC sampling, and results processing. Fundamentally, this
    class requires two objects to run, the Sampler object and the Data object, both of which can be obtained
    from the Acid object. If one or the other is not provided, some methods will not work."""

    def __init__(
            self,
            Acid_or_Data_or_Sampler,
            sampler                 : EnsembleSampler|None  = None,
            process_results         : bool                  = True,
            verbose                 : IntLike|bool|str|None = None,
        ) -> None:
        """Initiate the Result class

        Parameters
        ----------
        Acid_or_Data_or_Sampler : Acid | Data | emcee.EnsembleSampler
            An Acid object, Data object (contained in Acid class), or sampler object. If an Acid 
            object is provided, all other arguments are taken from there. If a Data object is 
            provided, a sampler can be provided in the second argument. If a sampler object 
            is provided, it will be used as the sampler, but all other attributes will need 
            to be set manually for the Result object to be fully functional.
        sampler : emcee.EnsembleSampler, optional
            A sampler object to use if the Data object was provided. If an Acid object 
            was provided, the sampler will be taken from there. If a sampler object was
            provided in the first argument, this will be ignored (with a warning), by default None
        process_results : bool, optional
            Whether to process the results from the Acid object upon initialisation, by default True.
            If False, the profiles attribute will not be available until Result.process_results() is called.
            The process_results functions does a LSD call, which can be skipped to save time and use
            the Result object for methods that do not require the profiles attribute, such as 
            continue_sampling() or plot_walkers(). This requires a Data object with the necessary attributes, 
            and a sampler object in the initialisation, or an Acid object with the necessary attributes already set.
            By default, None.
        verbose : int | bool | str, optional
            Verbosity level, works exactly the same as Acid verbosity, if not provided
            defaults to provided Acid class verbosity otherwise defaults to 2.
        # production_run : bool, optional
        #     Whether Acid was run in production mode, by default False
        """
        obj = Acid_or_Data_or_Sampler
        self.sampler    = None
        self.data       = None
        self.profiles = None
        self.config     = Config() # default config, will be updated if Acid or Data object is provided
        self.slurm      = "SLURM_JOB_ID" in os.environ

        self.config.verbose = verbose

        if hasattr(obj, "data") and hasattr(obj, "config") and hasattr(obj, "sampler"):
            # The above line is only all true if an Acid object (as they are set in initialisation), 
            # the sampler and data classes do not store all 3
            acid = obj
            self.initiate_data(acid.data)
            self.config = acid.config
            self.initiate_sampler(acid.sampler)
        elif isinstance(obj, Data):
            data = obj
            self.initiate_data(data)
            self.config = data.config
            if sampler is not None:
                self.initiate_sampler(sampler)
        elif isinstance(obj, EnsembleSampler):
            self.initiate_sampler(obj)
            if self.config.verbose>0:
                print("Warning: Data object not provided. Result object will not be fully functional.")
            return
        else:
            raise ValueError("First argument must be an Acid object, Data object, or emcee.EnsembleSampler object. "
                             f"Got {type(obj)} instead.")

        # From this point, a Data instance is provided and can be drawn from, but sampler may or may not be provided.
        # All frames must be available as a Result class variable due to legacy behaviour. Once created, we can point
        # Data.profiles to Result.profiles to keep them in sync.
        if process_results:
            self.process_results() # sets self.data.profiles, and points self.profiles to self.data.profiles
        else:
            if self.config.verbose>0:
                print("Warning: Results not processed. profiles attribute will not be available until " \
                "Result.process_results() is called.")

    @_require_data
    @_require_sampler
    def process_results(self) -> None:
        """Processes the MCMC sampler results to obtain the final LSD profiles and continuum fit, and errors on both."""
        t0 = time()

        # Obtain flattened samples
        flat_samples = self.sampler.get_chain(discard=self.burnin, thin=self.thin, flat=True)

        # Getting the final profile and continuum values
        nvel = len(self.data.velocities) if self.config.deterministic_profile is False else 0
        quartiles = np.percentile(flat_samples, [16, 50, 84], axis=0)
        errors = np.diff(quartiles, axis=0)
        errors = np.max(errors, axis=0) # why?
        self.poly_cos      = quartiles[1, nvel:]
        self.poly_cos_err  = errors[nvel:]  

        if self.config.verbose > 1:
            print('Getting the final profiles...')

        # Finding error for the continuum fit
        norm_wl = self.data.wavelengths["combined_normalized"]
        coeffs = flat_samples[:, nvel:]
        ncoeffs = coeffs.shape[1]
        powers = np.vander(norm_wl, N=ncoeffs, increasing=True)

        # First check memory to see if all samples can be used
        if self.slurm:
            available_memory = int(os.environ.get('SLURM_MEM_PER_NODE')) # in MB
            available_memory *= 1e6  # Convert to bytes as in the else statement below
        else:
            available_memory = psutil.virtual_memory().available
        m_available = available_memory * 1e-9 * 0.8 # in GB, with 0.8 factor safety gap
        n_samples, ncoeffs = coeffs.shape
        npix = powers.shape[0]
        matrix_size_gb = (2 * n_samples * npix + n_samples * ncoeffs + npix * ncoeffs) * 8 * 1e-9
        # If memory exceeded, fallback to using 1000 random samples
        if matrix_size_gb > m_available:
            if self.config.verbose > 1:
                print(f"Warning: Calculating continuum error with all samples may exceed available memory ({matrix_size_gb:.2f} GB required, {m_available:.2f} GB available). "
                "Calculating with a max of 1000 random samples instead.")
            indices_size = min(1000, n_samples)
            random_indices = np.random.choice(n_samples, size=indices_size, replace=False)
            coeffs = coeffs[random_indices, :]

        conts = (coeffs @ powers.T)
        continuum_error = np.std(conts, axis=0)

        # First get the combined profile, and then calculate each frame's profile if there are multiple frames.
        # If there is one frame, then the combined_profile is the same as the single frame profile.
        nframes = len(self.data.flux["input"])
        self.profiles = np.zeros((nframes, 2, len(self.data.velocities)))
        for counter in range(nframes+1):
            if counter == 0:
                flux = np.copy(self.data.flux["combined"])
                error = np.copy(self.data.errors["combined"])
                wavelengths = np.copy(self.data.wavelengths["combined"])
                sn = np.copy(self.data.sn["combined"])
                error[self.data.residual_masks] = 1e12
            else:
                flux = np.copy(self.data.flux["input"][counter-1])[self.data.nanmask]
                error = np.copy(self.data.errors["input"][counter-1])[self.data.nanmask]
                wavelengths = np.copy(self.data.wavelengths["input"][counter-1])[self.data.nanmask]
                sn = np.copy(self.data.sn["input"][counter-1])

                # Masking based off residuals interpolated onto new wavelength grid
                reference_wave = self.data.wavelengths["input"][np.nanargmax(self.data.sn["input"])]
                reference_wave = reference_wave[self.data.nanmask]

                reference_mask = np.zeros_like(reference_wave, dtype=bool)
                reference_mask[self.data.residual_masks] = True
                reference_interp1d = interp1d(reference_wave, reference_mask.astype(float), kind="nearest", bounds_error=False, fill_value=0.0)
                interpolated_mask = reference_interp1d(wavelengths) > 0.5
                error[interpolated_mask] = 1e12

            # Build continuum model
            a, b = utils.get_normalisation_coeffs(wavelengths)
            norm_wavelengths = (a*wavelengths)+b
            mdl = P.polyval(norm_wavelengths, self.poly_cos)

            # correcting continuum
            error = np.sqrt((error/mdl)**2 + (continuum_error/mdl)**2)
            flux /= mdl

            # Check whether we can skip alpha by reusing the same alpha, only true if the wavelength grid is identical
            condition = np.array_equal(wavelengths, self.data.wavelengths["combined"])
            alpha = self.data.alpha if condition else None

            LSD_profiles = LSD(self.data)
            LSD_profiles.run_LSD(wavelengths, flux, error, sn=sn, alpha=alpha)

            profile_f = LSD_profiles.profile_F
            profile_errors_f = LSD_profiles.profile_errors_F

            if counter == 0:
                self.combined_profile = [profile_f, profile_errors_f]
            else:
                self.profiles[counter-1] = [profile_f, profile_errors_f]

        self.data.profiles = self.profiles # point Data.profiles to Result.profiles to keep them in sync
        self.data.combined_profiles = self.combined_profile
        self.data.get_profiles_time = time() - t0
        self.data.full_run_time = self.data.initialisation_time + self.data.mcmc_time + self.data.get_profiles_time

        return

    @_require_profiles
    def __getitem__(self, item) -> np.ndarray:
        """Allows indexing into the profiles array directly from the Result object.
        """
        if isinstance(item, tuple):
            if len(item) == 3:
                # Legacy support for indexing style
                _order, frame, velocity = item
                item = (frame, velocity)
            elif len(item) == 2:
                item = item
            elif len(item) == 1:
                return self.combined_profile[item[0]]
        elif isinstance(item, int):
            return self.combined_profile[item]
        elif isinstance(item, str):
            if "combined" in item.lower():
                return self.combined_profile
            elif "profile" in item.lower():
                return self.profiles
        return self.profiles[item]

    @_require_profiles
    def __iter__(self) -> np.ndarray:
        """Allows iterating over the profiles array directly from the Result object.
        """
        return iter(self.profiles)

    def __repr__(self):
        # Only print out the sampler and data attributes, and whether profiles is available, to avoid printing large arrays
        return f"Result object with sampler={self.sampler}, data={self.data}, profiles={'available' if self.profiles is not None else 'not available'}"

    @_require_data
    @_require_sampler
    def continue_sampling(self, process_results:bool=True, sampler:EnsembleSampler|None=None, **kwargs) -> None:
        """Continue MCMC sampling for additional steps. Passes the stored sampler into a Acid instance with the saved data. See
        Acid.continue_sampling() for more details on the parameters that can be passed.

        Parameters
        ----------
        nsteps : int | None, optional
            Number of additional MCMC steps to run. Passed to Acid.continue_sampling with the stored sampler.
        max_steps : int | None, optional
            Maximum number of MCMC steps to run, by default None. Passed to Acid.continue_sampling with the stored sampler.
        max_steps_kwards : dict, optional
            Additional keyword arguments to be passed to the run_mcmc_until_converged function if max_steps is specified, by default None.
            The kwargs description can be found in Acid.ACID(), they are the 4 kwargs appearing after max_steps. Typos for kwargs are silently
            ignored. Passed to Acid.continue_sampling with the stored sampler.
        process_results : bool, optional
            Whether to process the results after continuing sampling, by default True.
            If False, the profiles attribute will not be updated until Result.process_results() is called.
        sampler : emcee.EnsembleSampler | None, optional
            Optionally provide a different sampler to continue sampling from, otherwise,
            takes the sampler from the Result object, by default None
        """
        if type(process_results) is int:
            raise ValueError("The process_results attribute must be a boolean, not an integer. Did you mean to set nsteps? If so, specificy nsteps=nsteps.")

        from .acid import Acid
        acid = Acid(data=self.data) # includes config data
        self.sampler = acid.continue_sampling(self.sampler, **kwargs)

        self.initiate_sampler(self.sampler) # update internal variables to match new sampler

        if process_results:
            self.process_results() # update profiles
        else:
            if self.config.verbose>0:
                print("Warning: Results not processed. profiles attribute will not be available until " \
                "Result.process_results() is called.")

    @_require_sampler
    def plot_walkers(
        self,
        sampler    : EnsembleSampler|None = None,
        burnin     : IntLike|None         = None,
        thin       : IntLike|None         = None,
        return_fig : bool                 = False
        ) -> None | tuple:
        """Plots, at maximum, the last 10 MCMC walkers for the LSD profile and continuum 
        polynomial coefficients.

        Parameters
        ----------
        sampler : emcee.EnsembleSampler, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        burnin : int | None, optional
            Optionally define the number of burnin steps, by default uses the burnin calculated when the sampler was initiated.
        thin : int | None, optional
            Optionally define the number of thinning steps, by default uses the thinning calculated when the sampler was initiated.
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False

        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing, else None
        """

        burnin = burnin if burnin is not None else self.burnin
        thin = thin if thin is not None else self.thin
        samples = self.sampler.get_chain(thin=int(thin))
        steps = np.arange(samples.shape[0]) * thin

        naxes = len(self.default_params)

        fig, ax = plt.subplots(naxes, 1, figsize=(10, 20), sharex=True)
        for i in range(naxes):
            ax[i].plot(steps, samples[:, :, self.default_params[i]], "k", alpha=0.3)
            ax[i].axvspan(0, burnin, color="red", alpha=0.1, label="burn-in")
            ax[i].set_ylabel(self.default_param_labels[i])
        ax[-1].legend()
        ax[-1].set_xlabel("Step number")
        ax[-1].set_xlim(0, self.nsteps)
        ax[0].set_title('MCMC Walkers')
        plt.subplots_adjust(hspace=0.05)
        if return_fig:
            return fig, ax
        plt.show()

    @_require_sampler
    def plot_corner(
        self,
        sampler    :EnsembleSampler|None = None,
        return_fig :bool                 = False,
        **kwargs,
        ) -> None | tuple:
        """Creates a corner plot for at maximum the last 8 LSD profile and continuum polynomial coefficients.

        Parameters
        ----------
        sampler : emcee.EnsembleSampler | None, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        return_fig : bool, optional
            Whether to return the figure object instead of showing the plot, by default False
        **kwargs:
            Additional keyword arguments to pass to corner.corner().
        
        Returns
        ----------
        If return_fig is True, returns the figure object containing the corner plot, else None
        """

        samples = self.sampler.get_chain()

        samples = self.sampler.get_chain(discard=self.burnin, flat=True, thin=self.thin)[:, self.default_params]

        fig = corner.corner(samples, labels=self.default_param_labels, show_title=True, title_fmt=".3f", title_kwargs={"fontsize": 16}, **kwargs)
        plt.suptitle('MCMC Corner Plot')
        if return_fig:
            return fig
        plt.show()

    @_require_profiles
    def plot_profiles(
        self,
        grid            :bool      = True,
        labels          :dict|None = None,
        return_fig      :bool      = False,
        subplot_kwargs  :dict|None = None,
        errorbar_kwargs :dict|None = None,
        fig_ax                     = None,
        ) -> None | tuple:
        """Plots the LSD profile result from Acid.

        Parameters
        ----------
        grid : bool, optional
            Show or hide grid, by default True
        labels : dict | None, optional
            Keys: 'xlabel', 'ylabel', and 'title'. Allows label overrides., by default None
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(), by default None
        errorbar_kwargs : dict | None, optional
            Keyword arguments to be passed to ax.errorbar(), by default None
        fig_ax : tuple[matplotlib.figure.Figure, matplotlib.axes.Axes] | None, optional
            Optionally provide an existing fig/axis tuple to plot on, by default None
        
        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing the plot.
        Otherwise, displays the plot and returns None.
        """
        # Set default errorbar kwargs
        errorbar_defaults = {
            "fmt"      : ".-",
            "ecolor"   : "red",
            "linewidth": 1,
        }
        errorbar_kwargs = utils.set_dict_defaults(errorbar_kwargs, errorbar_defaults)

        # Set default subplot kwargs
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 6)})

        # Set default labels
        default_labels = {
            "title" : "Acid Profile",
            "xlabel": "Velocity (km/s)",
            "ylabel": "Normalised Flux"
        }
        labels = utils.set_dict_defaults(labels, default_labels)

        # Set useful variables
        nframes = len(self.profiles)
        if fig_ax is None:
            fig, ax = plt.subplots(**subplot_kwargs)
        else:
            fig, ax = fig_ax

        for f, frame in enumerate(self.profiles):
            x, y, yerr = self.data.velocities, frame[0], frame[1]
            label_default = f"Frame {f+1}" if nframes > 1 else None
            # Override label in errorbar_kwargs if it is not already set, otherwise use the default label
            if "label" not in errorbar_kwargs:
                errorbar_kwargs["label"] = label_default
            ax.errorbar(x, y-1, yerr=yerr, **errorbar_kwargs)

        ax.set_title(labels["title"])
        ax.set_xlabel(labels["xlabel"])
        ax.set_ylabel(labels["ylabel"])
        ax.axhline(0, color='black', linestyle='--', linewidth=1)
        ax.legend()
        ax.grid(grid)
        if return_fig:
            return fig, ax
        else:
            plt.show()

    @_require_profiles
    def plot_forward_model(
        self,
        grid            :bool      = True,
        labels          :dict|None = None,
        return_fig      :bool      = False,
        subplot_kwargs  :dict|None = None,
        ) -> None | tuple:
        """Plots the forward model calculated from the final profiles to the combined input spectrum.

        Parameters
        ----------
        grid : bool, optional
            Show or hide grid, by default True
        labels : dict | None, optional
            Keys: 'xlabel', 'ylabel', 'title', and 'residuals_ylabel'. Allows label overrides, by default None
        return_fig : bool, optional
            Whether to return the figure and axis objects instead of showing the plot, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None
        
        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing the plot.
        Otherwise, displays the plot and returns None.
        """
        
        # Set default labels
        default_labels = {
            "title"           : "Forward Model Fit to Observed Spectrum",
            "xlabel"          : "Wavelength (Angstroms)",
            "ylabel"          : "Normalised Flux",
            "residuals_ylabel": "Residuals",
        }
        labels = utils.set_dict_defaults(labels, default_labels)

        # Set default subplot kwargs
        subplot_kwargs = {
            "figsize": (10, 8),
            "sharex": True,
            "gridspec_kw": {'height_ratios': [3, 1]}
        }
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 8)})

        # Get input data
        wavelengths = self.data.wavelengths["combined"]
        flux = self.data.flux["combined"]

        a, b = utils.get_normalisation_coeffs(wavelengths)
        profile = utils.flux_to_od(self.combined_profile[0])

        # Get flat_samples which are the same samples used to calculate the final profile
        continuum_model = P.polyval(wavelengths*a+b, self.poly_cos)
        model_flux = np.exp(- (self.data.alpha @ profile)) * continuum_model

        # Plotting
        fig, ax = plt.subplots(2, 1, **subplot_kwargs)
        ax[0].plot(wavelengths, flux, color='black', linewidth=1, label='Observed Spectrum')
        ax[0].plot(wavelengths, model_flux, color='C0', linewidth=1, label='Forward Model Fit')
        ax[0].plot(wavelengths, continuum_model, color='C1', linewidth=1, label='Fitted Continuum', linestyle='--')
        ax[1].plot(wavelengths, model_flux - flux, color='C0', linewidth=1, label='Residuals')
        ax[1].axhline(0, color='black', linestyle='--', linewidth=1)
        ax[0].set_title(labels["title"])
        ax[1].set_xlabel(labels["xlabel"])
        ax[0].set_ylabel(labels["ylabel"])
        ax[1].set_ylabel(labels["residuals_ylabel"])
        ax[1].axhline(0, color='black', linestyle='--', linewidth=1)
        ax[0].legend()
        ax[1].legend()
        ax[0].grid(grid)
        ax[1].grid(grid)
        plt.subplots_adjust(hspace=0.05)

        if return_fig:
            return fig, ax
        else:
            plt.show()

    @_require_sampler
    def plot_autocorrelation(
        self,
        sampler        : EnsembleSampler|None = None,
        burnin         : IntLike|None         = None,
        thin           : IntLike|None         = None,
        n_grid         : IntLike              = 12,
        c              : float                = 5.0,
        return_fig     : bool                 = False,
        subplot_kwargs : dict|None            = None,
        min_steps      : IntLike              = 100
        ) -> None | tuple:
        """
        Plot estimated integrated autocorrelation time as a function of chain length.

        From the emcee docs:
        - For several prefixes of the chain, estimate tau with Sokal windowing.
        - Plot tau(N) and the reference line tau = N/50.

        Parameters
        ----------
        sampler : emcee.EnsembleSampler | None, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        burnin, thin : int | None, optional
            Optional overrides. Defaults to self.burnin/self.thin from the sampler.
        n_grid : int, optional
            Number of N values (prefix lengths) to evaluate, by default 12.
        c : float, optional
            Sokal window constant, by default 5.0.
        return_fig : bool, optional
            Whether to return the figure and axes objects, by default False
        subplot_kwargs : dict | None, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None
        min_steps : int, optional
            Minimum number of post-burnin samples required to attempt autocorrelation estimation, by default 100
            If you decrease this, you may get unreliable estimates or errors from the autocorrelation time estimation.

        Returns
        ----------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing 
        the plot. Otherwise, displays the plot and returns None.
        """

        chain = self.sampler.get_chain()  # (nsteps, nwalkers, ndim)
        nsteps, nwalkers, ndim = chain.shape

        if nsteps < min_steps:
            raise ValueError("Not enough post-burnin samples to estimate autocorrelation reliably.")
        
        Ns = np.unique(np.exp(np.linspace(np.log(min_steps), np.log(nsteps), n_grid)).astype(int))
        Ns = Ns[Ns >= min_steps]  # Ensure we only consider N >= min_steps

        tau_estimates = {p: np.full(len(Ns), np.nan, dtype=float) for p in self.default_params}

        # Estimate taus
        for i, n in enumerate(Ns):
            for p in self.default_params:
                y = chain[:n, :, p].T
                tau_estimates[p][i] = utils.autocorr_new(y, c=c)

        subplot_kwargs = {} if subplot_kwargs is None else dict(subplot_kwargs)
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 6)})

        fig, ax = plt.subplots(**subplot_kwargs)

        for label, p in zip(self.default_param_labels, self.default_params):
            ax.loglog(Ns, tau_estimates[p], "o-", label=f"{label}")

        # Reference line tau = N/50
        ax.loglog(Ns, Ns / 50.0, "--", label=r"$\tau = N/50$")

        ax.set_xlabel("number of post-burnin samples per walker (N)")
        ax.set_ylabel(r"estimated integrated autocorrelation time $\tau$")
        ax.set_title("Autocorrelation time estimates vs chain length")
        ax.legend()
        ax.grid(True, which="both")

        if return_fig:
            return fig, ax
        plt.show()

        return

    @_require_sampler
    def plot_acf(
        self,
        sampler        : EnsembleSampler|None = None,
        max_lag        : IntLike|None         = None,
        return_fig     : bool                 = False,
        subplot_kwargs : dict|None            = None,
        ) -> None | tuple:
        """
        Plot the autocorrelation function (ACF) for each parameter, averaged across walkers.
        
        Parameters
        ----------
        sampler : emcee.EnsembleSampler, optional
            Optionally provide a different sampler to plot from, otherwise,
            takes the sampler from the Result object, by default None
        max_lag : int, optional
            Maximum lag to plot, by default None (plots up to min(5000, nsteps-1))
        return_fig : bool, optional
            Whether to return the figure and axes objects, by default False
        subplot_kwargs : dict, optional
            Keyword arguments to be passed to plt.subplots(). Allows label overrides, by default None

        Returns
        -------
        If return_fig is True, returns a tuple (fig, ax) of the figure and axes objects containing 
        the plot. Otherwise, displays the plot and returns None.
        """
        chain = self.sampler.get_chain() 
        nsteps, nwalkers, ndim = chain.shape

        subplot_kwargs = {} if subplot_kwargs is None else dict(subplot_kwargs)
        subplot_kwargs = utils.set_dict_defaults(subplot_kwargs, {"figsize": (10, 5)})

        fig, ax = plt.subplots(**subplot_kwargs)

        for param, label in zip(self.default_params, self.default_param_labels):

            y = chain[:, :, param].T  # (nwalkers, nsteps)

            # Mean ACF across walkers
            f = np.zeros(nsteps)
            for w in range(nwalkers):
                f += utils.autocorr_func_1d(y[w], norm=True)
            f /= nwalkers

            if max_lag is None:
                max_lag = min(5_000, nsteps - 1)
            max_lag = int(max_lag)

            ax.plot(np.arange(max_lag + 1), f[: max_lag + 1], label=f"{label}")
        
        ax.set_xlabel("Lag (steps)")
        ax.set_ylabel("Autocorrelation")
        ax.set_title(f"Mean ACF across walkers")
        ax.set_xscale("log")
        ax.grid(True)
        ax.axhline(0, color="black", linestyle="--", linewidth=1)
        ax.legend()

        if return_fig:
            return fig, ax
        plt.show()

    def initiate_sampler(self, sampler:EnsembleSampler|None) -> None:
        """Initiates the sampler attribute from an external sampler.

        Parameters
        ----------
        sampler : emcee.EnsembleSampler
            An emcee EnsembleSampler object to set as the sampler attribute.
        """
        self.sampler = sampler if sampler is not None else self.sampler
        if self.sampler is None:
            raise ValueError("A sampler must be provided in initialisation or in method call")
        if sampler is None:
            return # sampler already initiated from initialisation, so skip the rest of the method

        self.ndim = self.sampler.ndim
        self.nwalkers = self.sampler.nwalkers
        self.nsteps = self.sampler.get_chain().shape[0]

        # Calculate autocorr time, burnin, thin
        # Suppress output from get_autocorr_time call
        with open(os.devnull, "w") as devnull, \
            contextlib.redirect_stdout(devnull), \
            contextlib.redirect_stderr(devnull):
            self.tau = self.sampler.get_autocorr_time(quiet=True)
        
        self.converged = True
        if self.nsteps < 50 * np.max(self.tau):
            self.converged = False
            if self.config.verbose>1:
                print("The number of MCMC steps is less than 50 times the maximum autocorrelation " \
                "time.\n The sampler may not have converged. Consider running more steps or checking " \
                f"the walker plots.\n The max autocorrelation time is {np.max(self.tau):.2f}, therefore " \
                f"the minimum number of steps should be roughly {int(50 * np.max(self.tau))}.\n Disabling burnin " \
                f"from autocorrelation time, instead using burnin=steps-1000")

        try:
            self.thin = int(np.min(self.tau)/5)
            if self.converged:
                self.burnin = int(3 * np.max(self.tau))
            else:
                self.burnin = self.nsteps - 1000 # just the last 1000 steps
        except:
            if self.config.verbose>0:
                print(f"Warning: Could not compute autocorrelation time for burnin and thinning.\n This is likely" \
                f" due to all posterior samples being rejected by prior constraints.\n The resulting profile is likely" \
                f" wrong. Setting burnin=nsteps-1000, and thin=1.")
            self.burnin = self.nsteps - 1000 # just the last 1000 steps
            self.thin = 1

        if self.config is not None:
            deterministic = self.config.deterministic_profile
            n_poly_params = self.data.config.poly_ord + 1
        else: # Make a best guess
            if self.ndim > 6: # ie we assume a poly order of 5 is the highest anyone would ever want to go
                deterministic = False
                n_poly_params = 4
            else:
                deterministic = True
                n_poly_params = self.ndim

        poly_params = np.arange(-1, -n_poly_params-1, -1).tolist()
        a=ord('a')
        alph=[chr(i) for i in range(a,a+26)]
        poly_labels = [alph[i] for i in range(n_poly_params)]
        
        samples = self.sampler.get_chain(thin=int(self.thin), discard=int(self.burnin))
        if not deterministic:
            max_profile_idx = np.argmax(samples[:,:,:-n_poly_params].mean(axis=(0,1)))
            poly_params.extend([-5, max_profile_idx, 1])
            poly_labels.extend(["$Z_{-1}$", "$Z_{max}$", "$Z_0$"])
        
        self.default_params = poly_params
        self.default_param_labels = poly_labels

    def initiate_data(self, data:Data) -> None:
        """Initiates the data attribute from an external Data object.

        Parameters
        ----------
        data : Data
            A Data object to set as the data attribute.
        """
        self.data = data if data is not None else getattr(self, "data", None)
        if self.data is None:
            raise ValueError("A Data object must be provided in initialisation or in method call")
        if data is None:
            return # data already initiated from initialisation, so skip the rest of the method

        self.profiles = self.data.profiles
        self.nsteps     = self.data.nsteps

        # For convenience, let the user call the model without needing to input all required args
        MCMC_class = mcmc.MCMC(self.data)
        self.model = MCMC_class.run_model_function

    @_require_data
    @_require_sampler # TODO check if necessary
    def save_result(self, filename:str="result.pkl", store_sampler:bool=True) -> None:
        """Saves the Result object to a pickle file.

        Parameters
        ----------
        filename : str, optional
            Name of the file to save the Result object to, by default "result.pkl"
        store_sampler : bool, optional
            Whether to store the sampler backend in the pickle file. If False, 
            the sampler will not be stored, and the Result object will not be able to 
            continue sampling or plot walkers/corner plots
        """
        state = dict(self.__dict__)

        state["data"] = self.data.to_dict()
        state["backend"] = dict(self.sampler.backend.__dict__) if store_sampler else None
        state["model"] = None
        state["sampler"] = None

        with open(filename, "wb") as f:
            pickle.dump(state, f, protocol=pickle.HIGHEST_PROTOCOL)

        if getattr(self, "config", None) is not None and self.config.verbose > 1:
            print(f"Result object saved to {filename}")

    @classmethod
    def load_result(cls, result_object:str|object="result.pkl"):
        """Loads a Result object from a pickle file or from an object with the same attributes as a saved Result object.

        Parameters
        ----------
        result_object : str | object, optional
            A pickle file name or an object with the same attributes as a saved Result object, by default "result.pkl"

        Returns
        ----------
        Result
            A Result object loaded from the pickle file or from the provided object.
        """
        if isinstance(result_object, str):
            with open(result_object, "rb") as f:
                obj = pickle.load(f)
        else:
            obj = dict(result_object.__dict__)

        res = cls.__new__(cls) 
        res.__dict__.update(obj)

        # reconstruct data
        data = Data()
        data.from_dict(res.data)
        res.data = data

        # reconstruct backend
        if res.backend is not None:
            backend = emceebackend.Backend(dtype=np.float64)
            backend.__dict__.update(res.backend)

            # reconstruct sampler from backend
            shape = backend.shape
            log_prob = mcmc.MCMC(res.data)
            res.sampler = EnsembleSampler(*shape, log_prob, backend=backend) # dummy sampler to hold the backend
            res.backend = None # backend is now stored in the sampler, so remove it from the Result object to avoid confusion

        # rebuild convenience things that shouldn’t be pickled
        cls.initiate_data(res, res.data) # sets profiles and nsteps

        if getattr(res, "sampler", None) is not None:
            cls.initiate_sampler(res, res.sampler) # sets burnin, thin, and default params/labels
        
        if getattr(res, "config", None) is None:
            res.config = Config()
        
        if res.config.verbose > 1:
            print("Result object loaded")

        return res


from __future__ import annotations
import numpy as np
from astropy.io import  fits
import glob, psutil, os
import scipy.constants as const
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
from scipy.interpolate import LSQUnivariateSpline
from tqdm import tqdm
from numpy import integer as npint
from scipy.linalg import cho_factor, cho_solve
from beartype import beartype
from . import utils
from .data import Config
from .utils import c_kms, IntLike, Scalar, Array1D, Array2D

@beartype
class LSD:
    """Class containing all useful functions for performing least-squares deconvolution.
    This does not simultaneously fit continuum and perform LSD (which ACID does). It is used
    for the initial parameters of the ACID mcmc run and for obtaining final profiles. It 
    can also be used as a standalone LSD implementation.
    """
    def __init__(self, Data:object|None=None):
        """Initialises the LSD class, optionally with a Data instance to take parameters from.

        Parameters
        ----------
        Data : object | None, optional
            A data instance to draw parameters and configs from, by default None
        """
        self.slurm            = "SLURM_JOB_ID" in os.environ
        self.data             = Data if Data is not None else None
        self.linelist         = Data.linelist if Data is not None else None

        try:
            self.config = Data.config
        except:
            self.config = Config() # uses defaults

    def run_LSD(
        self,
        wavelengths : Array1D,
        flux        : Array1D,
        errors      : Array1D,
        sn          : Scalar|Array1D,
        linelist                   = None,
        velocities  : Array1D|None = None,
        verbose     : IntLike|None = None,
        alpha       : Array2D|None = None,
        ):
        """Runs the LSD algorithm to extract the average line profile from the observed spectrum.

        Parameters
        ----------
        wavelengths : np.ndarray
            Array of wavelengths of the observed spectrum in Angstroms
        flux : np.ndarray
            Array of flux values corresponding to the wavelengths (in linear space, and should be continuum normalized)
        errors : np.ndarray
            Array of error values corresponding to the flux
        sn : float | int
            Signal-to-noise ratio of the observed spectrum
        linelist : str | dict, optional
            Path to the linelist file or a dictionary containing 'wavelengths' and 'depths'. If the class was 
            not initialised with an Acid instance, this is required, by default None
        velocities : np.ndarray, optional
            Array of velocities corresponding to the observed spectrum. If the class was not initialised with 
            an Acid instance, this is required, by default None
        verbose : int | None, optional
            Verbosity level, if None, uses the class default of 2. See the Acid class for more information about
            verbosity integer levels, by default None
        alpha : np.ndarray, optional
            Precomputed alpha matrix, if already calculated and you want to skip directly to the Cholesky 
            decomposition and solving for the profile, by default None
        """

        if not wavelengths.shape == flux.shape == errors.shape:
            raise ValueError("Input wavelengths, flux, and errors must have the same shape.")
        if wavelengths.ndim != 1:
            raise ValueError("Input wavelengths, flux, and errors must be 1D arrays.")        

        self.data.velocities = velocities if velocities is not None else self.data.velocities
        self.config.update_hipri(verbose=verbose, linelist=linelist) # Update config with new values, if not None, otherwise keep old values

        # Unpack the linelist stored in data
        wavelengths_linelist, depths_linelist = self.data.linelist

        # Clip linelist to wavelength range of spectrum
        wavelengths_linelist, depths_linelist = utils.clip_wavelengths(wavelengths, wavelengths_linelist, depths_linelist)
        if len(wavelengths_linelist) == 0:
            raise ValueError(f"No lines in linelist are within the wavelength range of the observed spectrum. \n"\
                             f"You may have mismatched wavelengths units between linelist and spectrum or an empty linelist.\n"\
                             f"Please check your linelist and input spectrum.")

        # Apply S/N cut (of 1/(3*SN)) to linelist
        wavelengths_linelist, depths_linelist = self.sn_clip(wavelengths_linelist, depths_linelist, sn)

        # Convert to optical depth space for the linelist and the spectrum (may move to own function)
        flux, errors, depths_linelist = utils.flux_to_od(flux, errors, depths_linelist)

        # Calculates alpha in optical depth, selects lines greater than 1/(3*sn)
        if alpha is None:
            self.alpha = self.calc_alpha(wavelengths, wavelengths_linelist, depths_linelist)
        else:
            self.alpha = alpha

        # Now solve for profile using Cholesky decomposition
        self.c_factor = self.calc_cholesky(self.alpha, errors)

        # Solve for profile and profile errors using Cholesky factors
        self.profile, self.profile_errors = self.solve_z(self.alpha, flux, errors, self.c_factor)

        # Convert profile back to flux if needed
        self.profile_F, self.profile_errors_F = utils.od_to_flux(self.profile, self.profile_errors)

        return

    def sn_clip(self, wavelengths_linelist, depths_linelist, sn):
        """Applies a signal-to-noise cut to the linelist, removing lines shallower than 1/(3*sn).

        Parameters
        ----------
        wavelengths_linelist : np.ndarray
            Wavelengths from the linelist
        depths_linelist : np.ndarray
            Depths from the linelist
        sn : float
            Signal-to-noise ratio threshold

        Returns
        -------
        np.ndarray
            Clipped wavelengths from the linelist
        np.ndarray
            Clipped depths from the linelist
        """
        # Selecting lines deeper than 1/(3*sn)
        idx = (depths_linelist >= 1/(3*sn))
        wavelengths_linelist = wavelengths_linelist[idx]
        depths_linelist = depths_linelist[idx]

        # Analyse remaining lines
        ncut = np.sum(~idx)
        nrest = np.sum(idx)
        perc = 100 * nrest / (nrest + ncut)
        if nrest == 0:
            raise ValueError(f"No lines remain in the linelist after S/N cut. Please check your linelist and S/N value.")
        if self.config.verbose > 0:
            if perc < 5:
                print("Warning: Less than 5% of lines remain after S/N cut. Check your linelist and S/N value.")
            if self.config.verbose > 2 or perc < 5:
                print(f"{perc:.2f}% of lines used in LSD: {nrest} out of {nrest + ncut} remain from S/N cut.")
        return wavelengths_linelist, depths_linelist

    def calc_alpha(
        self,
        wavelengths          : Array1D,
        wavelengths_linelist : Array1D,
        depths_linelist      : Array1D,
        velocities           : Array1D|None = None,
        verbose              : IntLike|None = None,
        ):
        """Calculates the alpha matrix given flux and errors in OD space, and a line_list in OD space.
        Note that if this function is called without using run_LSD, there is no selection of lines deeper than 1/(3*sn).
        If you still wish to do this, it needs to be done in linear space with the sn_clip function before converting to OD space.

        Parameters
        ----------
        wavelengths : np.ndarray
            Array of wavelengths of the observed spectrum in optical depth space
        wavelengths_linelist : np.ndarray
            Array of wavelengths from the linelist in optical depth space
        depths_linelist : np.ndarray
            Array of depths from the linelist in optical depth space
        velocities : np.ndarray, optional
            Array of velocities, needs to either be initialised by class with Acid instance, or input here, by default None
        verbose : int | None, optional
            Verbosity level, uses the class default of 2 if None, by default None
        """

        self.data.velocities = velocities if velocities is not None else self.data.velocities
        self.config.update_hipri(verbose=verbose) # Update config with new values, if not None, otherwise keep old values

        # Calculate velocity pixel size
        deltav = self.data.velocities[1] - self.data.velocities[0]

        # Clip linelist to wavelength range of spectrum (again just in case this is called without run_LSD)
        wavelengths_linelist, depths_linelist = utils.clip_wavelengths(wavelengths, wavelengths_linelist, depths_linelist)

        # Find differences and velocities
        blankwaves = wavelengths
        diff = blankwaves[:, np.newaxis] - wavelengths_linelist
        vel = c_kms * (diff / wavelengths_linelist)

        if self.slurm:
            available_memory = int(os.environ.get('SLURM_MEM_PER_NODE')) # in MB
            available_memory *= 1e6  # Convert to bytes as in the else statement below
        else:
            available_memory = psutil.virtual_memory().available

        mat_size = len(wavelengths_linelist) * len(self.data.velocities) * len(blankwaves) * 8 * 1e-9 # Matrix size in GB
        m_available = available_memory * 1e-9 / 2  # Available memory in GB (divided by 2 to be safe)

        # We can calculate the alpha matrix in one pass if the number of wavelengths is small enough
        if mat_size < m_available:
            # Calculating entire alpha matrix at once
            x = (vel[:, :, np.newaxis] - self.data.velocities) / deltav
            delta = np.clip(1.0 - np.abs(x), 0.0, 1.0)
            alpha = (depths_linelist[:, None] * delta).sum(axis=1)  # (nb, n_vel)

        # Else, calculate in blocks to save memory
        else:
            n_blank = len(blankwaves)
            n_vel   = len(self.data.velocities)
            mem_size = available_memory // 2
            bytes_per_row = n_blank * n_vel * 8 * 3 # *8 for float64, *3 for vel, x, delta in a row
            max_block = max(1, mem_size // bytes_per_row)
            block = int(min(max_block, len(wavelengths_linelist)))
            # Set initial alpha matrix to np.zeros
            alpha  = np.zeros((len(blankwaves), len(self.data.velocities)), dtype=np.float64)

            # Use tqdm progress bar if verbose
            if self.config.verbose>1:
                iterable = tqdm(range(0, len(wavelengths_linelist), block), desc='Calculating alpha matrix')
            else:
                iterable = range(0, len(wavelengths_linelist), block)

            for start_pos in iterable:
                # Ensure we don't go out of bounds on last iteration
                end_pos = min(start_pos + block, len(wavelengths_linelist))
                wl  = wavelengths_linelist[start_pos:end_pos]
                dep = depths_linelist[start_pos:end_pos]

                # Perform calculations for this block
                vel_blk = c_kms * (blankwaves[:, None] - wl) / wl
                x_blk   = (vel_blk[:, :, None] - self.data.velocities) / deltav
                delta   = np.clip(1.0 - np.abs(x_blk), 0.0, 1.0)                    

                alpha += (dep[:, None] * delta).sum(axis=1)
        return alpha

    @staticmethod
    def calc_cholesky(
        alpha : Array2D,
        error : Array1D,
        **kwargs,
        ):
        """Calculates the LHS Cholesky factorisation matrix given the alpha matrix and flux errors (in optical depth space)


        Parameters
        ----------
        alpha : np.ndarray
            The precomputed alpha matrix
        error : np.ndarray
            Flux errors in optical depth space
        **kwargs : dict, optional
            Additional keyword arguments to pass to scipy.linalg.cho_factor. 
            Overwrite_a=False is already set by default, do not pass this as a kwarg.

        Returns
        -------
        c_factor : tuple
            Cholesky factorisation matrix and lower/upper flag, to be put straight into solve_z as c_factor
        """
        V = 1.0 / (error ** 2) # variance vector in log space, error already in log space

        # M = αT V α,  b = αT V R
        AVA = alpha.T @ (V[:, None] * alpha)

        # Diangostics for common 1-th leading order linalg error
        # M = alpha.T @ (V[:, None] * alpha)
        # print("finite M:", np.all(np.isfinite(M)))
        # print("min diag:", np.min(np.diag(M)))
        # print("rank:", np.linalg.matrix_rank(M), " / ", M.shape[0])
        # col_norm = np.linalg.norm(np.sqrt(V)[:, None] * alpha, axis=0)
        # print("zero columns:", np.sum(col_norm == 0))

        # Cholesky factorisation of M
        c_factor = cho_factor(AVA, overwrite_a=False, **kwargs)
        return c_factor

    @staticmethod
    def solve_z(
        alpha,
        flux,
        error,
        c_factor,
        return_error : bool = True,
        return_cov : bool = False,
        **kwargs,
        ):
        """Solves for the LSD profile and its errors using the Cholesky factors.

        Parameters
        ----------
        alpha : np.ndarray
            The precomputed alpha matrix
        flux : np.ndarray
            The observed flux values in optical depth space
        error : np.ndarray
            The flux errors in optical depth space
        c_factor : tuple
            Cholesky factorisation matrix and lower/upper flag, to be put straight into 
            scipy.linalg.cho_solve as c_factor
        return_error : bool, optional
            Whether to calculate and return the profile errors along with the
            profile, by default True
        return_cov : bool, optional
            Whether to return the full covariance matrix instead of just the errors, by default False
        **kwargs : dict, optional
            Additional keyword arguments to pass to both scipy.linalg.cho_solve calls
            (one for the profile, one for the covariance matrix)
            Overwrite_b=False is already set by default, do not pass this as a kwarg.

        Returns
        -------
        profile, profile_errors, cov_z : tuple
            LSD profile and its errors (if return_error is True) and covariance matrix (if return_cov is True)
        """
        V = 1.0 / (error ** 2) # variance vector in log space, error already in log space
        R = flux         # R matrix in log space

        # M = αT V α,  b = αT V R
        AVR = alpha.T @ (V * R)

        # z = M⁻¹ b
        profile = cho_solve(c_factor, AVR, overwrite_b=False, **kwargs)

        # Find error, cov(z) = M⁻¹, take diagonal
        if return_error or return_cov:
            AVA = alpha.T @ (V[:, None] * alpha)
            cov_z = cho_solve(c_factor, np.eye(AVA.shape[0]), overwrite_b=False, **kwargs)
            profile_errors = np.sqrt(np.diag(cov_z))
            if return_cov:
                return profile, profile_errors, cov_z
            else:
                return profile, profile_errors
        else:
            return profile

    def get_wave(self, data, header):

        wave = np.array(data*0., dtype = 'longdouble')
        no = data.shape[0]
        npix = data.shape[1]
        d = header['ESO DRS CAL TH DEG LL']
        xx0 = np.arange(npix)
        xx = []
        for i in range(d+1):
            xx.append(xx0**i)
        xx = np.asarray(xx, dtype = 'longdouble')

        for o in range(no):
            for i in range(d+1):
                idx = i + o * (d + 1)
                par = np.longdouble(header['ESO DRS CAL TH COEFF LL%d' % idx])
                wave[o, :] = wave[o, :] + par * xx[i, :]
            #for x in range(npix):
            #  wave[o,x]=wave[o,x]+par*xx[i,x]#float(x)**float(i)

        return wave

    def upper_envelope(self, x, y):
        # from MM-LSD code - give credit if needed
        # used to compute the tapas continuum. find peaks then fit spline to it.
        peaks = find_peaks(y, height=0.2, distance=len(x) // 500)[0]
        # t= knot positions
        spl = LSQUnivariateSpline(x=x[peaks], y=y[peaks], t=x[peaks][5::10])
        return spl(x)

    def blaze_correct(self, file_type, spec_type, order, file, directory, masking, run_name, berv_opt):
        #### Inputing spectrum depending on file_type and spec_type #####

        if file_type == 's1d':
            #### Finding min and max wavelength from e2ds for the specified order ######
            file_e2ds = file.replace('s1d', 'e2ds')

            hdu=fits.open('%s'%file_e2ds)
            sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]
            spec = hdu[0].data
            header = hdu[0].header
            brv = header['ESO DRS BERV']

            spec_check = spec[spec<=0]
            # if len(spec_check)>0:
                # print('WARNING NEGATIVE/ZERO FLUX - corrected')

            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)

            wave=self.get_wave(spec, header)*(1.+brv/2.99792458e5)
            wavelengths_order = wave[order]
            wavelength_min = np.min(wavelengths_order)
            wavelength_max = np.max(wavelengths_order)

            ## remove overlapping region (always remove the overlap at the start of the order, i.e the min_overlap)
            last_wavelengths = wave[order-1]
            next_wavelengths = wave[order+1]
            min_overlap = np.max(last_wavelengths)
            max_overlap = np.min(next_wavelengths)

            # idx_ = tuple([wavelengths>min_overlap])
            # wavelength_min = 5900
            # wavelength_max = wavelength_min+200    ###### if you want to do a WAVELENGTH RANGE just input min and max here ######
            hdu.close()

            #### Now reading in s1d file ########
            hdu=fits.open('%s'%file)
            spec=hdu[0].data
            header=hdu[0].header
            spec_check = spec[spec<=0]
            
            wave=hdu[0].header['CRVAL1']+(np.arange(spec.shape[0]))*hdu[0].header['CDELT1']
            
            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)

            if spec_type == 'order':
                wavelengths=[]
                fluxes=[]
                errors = []
                for value in range(0, len(wave)):
                    l_wave=wave[value]
                    l_flux=spec[value]
                    l_error=flux_error[value]
                    if l_wave>=wavelength_min and l_wave<=wavelength_max:
                        wavelengths.append(l_wave)
                        fluxes.append(l_flux)
                        errors.append(l_error)
                wavelengths = np.array(wavelengths)
                fluxes = np.array(fluxes)
                errors = np.array(errors)


                if len(wavelengths)>5144:
                    wavelengths = wavelengths[:5144]
                    fluxes = fluxes[:5144]
                    flux_error = errors[:5144]

                spec_check = fluxes[fluxes<=0]
                # if len(spec_check)>0:
                #     print('WARNING NEGATIVE/ZERO FLUX - corrected')

                # where_are_zeros = (spec<=0)
                # spec[where_are_zeros] = 1e12
                # flux_error = np.sqrt(spec)
                
                flux_error_order = flux_error
                masked_waves = []
                masked_waves = np.array(masked_waves)

                if masking == 'masked':
                    ## I've just been updating as I go through so it's not complete
                    #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
                    masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
                    min_waves_mask = np.array(masks_csv[:,0])
                    max_waves_mask = np.array(masks_csv[:,1])
                    masks = []
                    for mask_no in range(len(min_waves_mask)):
                        masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

                    masked_waves=[]

                    for mask in masks:
                        #print(np.max(mask), np.min(mask))
                        idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                        #print(flux_error_order[idx])
                        flux_error_order[idx] = 1e12
                        #print(idx)
                        if len(wavelengths[idx])>0:
                            masked_waves.append(wavelengths[idx])

                    #masks = []
                    ### allows extra masking to be added ##

                    plt.figure('masking')
                    plt.plot(wavelengths, fluxes)
                    if len(masked_waves)>0:
                        for masked_wave in masked_waves:
                            plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                    #plt.show()

                    #response = input('Are there any regions to be masked? y or n: ')
                    response = 'y'
                    if response == 'y':
                        '''
                        print('Take note of regions.')
                        plt.figure('masking')
                        plt.plot(wavelengths, fluxes)
                        plt.show()
                        '''
                        #response1 = int(input('How many regions to mask?: '))
                        response1 = 0
                        for i in range(response1):
                            min_wave = float(input('Minimum wavelength of region %s: '%i))
                            max_wave = float(input('Maximum wavelength of region %s: '%i))
                            masks.append([min_wave, max_wave])
                        masked_waves=[]
                        #print(masks)
                        for mask in masks:

                            idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))

                            flux_error_order[idx] = 1e12

                            if len(wavelengths[idx])>0:
                                masked_waves.append(wavelengths[idx])

                        plt.figure('telluric masking')
                        plt.title('Spectrum - after telluric masking')
                        plt.plot(wavelengths, fluxes)
                        for masked_wave in masked_waves:
                            plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                        plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))

                        plt.figure('errors')
                        plt.plot(wavelengths, flux_error_order)
                        plt.close('all')
                        #plt.show()

                elif masking == 'unmasked':
                    masked_waves = []
                    masked_waves = np.array(masked_waves)

            elif spec_type == 'full':
                ## not set up properly.
                wavelengths = wave
                fluxes = spec

        elif file_type == 'e2ds':
            hdu=fits.open('%s'%file)
            spec=hdu[0].data
            header=hdu[0].header
            sn = hdu[0].header['HIERARCH ESO DRS SPE EXT SN%s'%order]

            spec_check = spec[spec<=0]
            # if len(spec_check)>0:
            #     print('WARNING NEGATIVE/ZERO FLUX - corrected')


            # where_are_NaNs = np.isnan(flux_error)
            # flux_error[where_are_NaNs] = 1e12
            where_are_zeros = (spec<=0)
            spec[where_are_zeros] = 1e12
            flux_error = np.sqrt(spec)
            '''
            flux_error1 = header['HIERARCH ESO DRS SPE EXT SN%s'%order]

            flux_error = header['HIERARCH ESO DRS CAL TH RMS ORDER%s'%order]
            print(flux_error, flux_error1)

            flux_error = flux_error*np.ones(np.shape(spec))
            '''
            # file_ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
            # print(file_ccf[0].header['ESO DRS BERV'])
            brv = np.longdouble(header['ESO DRS BERV'])
            # print(brv)
            wave_nonad = self.get_wave(spec, header)
            # if berv_opt == 'y':
            #     print('BERV corrected')
            wave = wave_nonad*(1.+brv/2.99792458e5)
            wave = np.array(wave, dtype = 'double')
            # if berv_opt == 'n':
            #     print('BERV not corrected')
            # wave = wave_nonad
        
            # rv_drift=header['ESO DRS DRIFT RV'] 
            # print(rv_drift)
            wave_corr = (1.+brv/2.99792458e5)
            # print(brv, (wave_corr-1)*2.99792458e5)

            # inp = input('Enter to continue...')
            '''
            plt.figure('Spectrum directly from fits file')
            plt.title('Spectrum directly from fits file')
            plt.errorbar(wave[order], spec[order], yerr = flux_error[order])
            plt.xlabel('wavelengths')
            plt.ylabel('flux')
            plt.show()
            '''

            blaze_file = glob.glob('tests/data/*blaze_B*.fits')

            blaze_file = blaze_file[0]
            # try:
            #     blaze_file = glob.glob('./**blaze_A*.fits')
            #     # print('%sblaze_folder/**blaze_A*.fits'%(directory))
            #     # print(blaze_file)
            #     blaze_file = blaze_file[0]
            # except: 
            #     try:
            #         blaze_file = glob.glob('/Users/lucydolan/Starbase/problem_frames/**blaze_A*.fits')
            #         blaze_file = blaze_file[0]
            #     except:
            #         blaze_file = glob.glob('tests/data/**blaze_A*.fits')
            blaze = fits.open('%s'%blaze_file)
            blaze_func = blaze[0].data
            min_rows = min(spec.shape[0], blaze_func.shape[0], flux_error.shape[0])
            spec, blaze_func, flux_error = spec[:min_rows], blaze_func[:min_rows], flux_error[:min_rows]
            spec = spec/blaze_func
            flux_error = flux_error/blaze_func

            fluxes = spec[order]
            flux_error_order = flux_error[order]
            wavelengths = wave[order]

            if masking == 'masked':
                ## I've just been updating as I go through so it's not complete
                #if you want to add to it the just add an element of form: [min wavelength of masked region, max wavelength of masked region]
                masks_csv = np.genfromtxt('/home/lsd/Documents/HD189733b_masks.csv', delimiter=',')
                min_waves_mask = np.array(masks_csv[:,0])
                max_waves_mask = np.array(masks_csv[:,1])
                masks = []
                for mask_no in range(len(min_waves_mask)):
                    masks.append([min_waves_mask[mask_no], max_waves_mask[mask_no]])

                masked_waves=[]

                for mask in masks:
                    #print(np.max(mask), np.min(mask))
                    idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                    #print(flux_error_order[idx])
                    flux_error_order[idx] = 1e12
                    #print(idx)
                    if len(wavelengths[idx])>0:
                        masked_waves.append(wavelengths[idx])

                #masks = []
                ### allows extra masking to be added ##

                plt.figure('masking')
                plt.plot(wavelengths, fluxes)
                if len(masked_waves)>0:
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                #plt.show()

                #response = input('Are there any regions to be masked? y or n: ')
                response = 'y'
                if response == 'y':
                    '''
                    print('Take note of regions.')
                    plt.figure('masking')
                    plt.plot(wavelengths, fluxes)
                    plt.show()
                    '''
                    #response1 = int(input('How many regions to mask?: '))
                    response1 = 0
                    for i in range(response1):
                        min_wave = float(input('Minimum wavelength of region %s: '%i))
                        max_wave = float(input('Maximum wavelength of region %s: '%i))
                        masks.append([min_wave, max_wave])
                    masked_waves=[]
                    #print(masks)
                    for mask in masks:
                        #print(np.max(mask), np.min(mask))
                        idx = np.logical_and(wavelengths>=np.min(mask), wavelengths<=np.max(mask))
                        #print(flux_error_order[idx])
                        flux_error_order[idx] = 1e12

                        if len(wavelengths[idx])>0:
                            masked_waves.append(wavelengths[idx])

                    plt.figure('Telluric masking')
                    plt.title('Spectrum - telluric masking')
                    plt.plot(wavelengths, fluxes)
                    # print(masked_waves)
                    for masked_wave in masked_waves:
                        plt.axvspan(np.min(masked_wave), np.max(masked_wave), alpha=0.5, color='red')
                    #print('new version')
                    plt.savefig('/home/lsd/Documents/LSD_Figures/masking_plots/order%s_masks_%s'%(order, run_name))
                    plt.ylabel('flux')
                    plt.xlabel('wavelength')

                if response == 'n':
                    print('yay!')


            elif masking == 'unmasked':
                masked_waves = []
                masked_waves = np.array(masked_waves)

            else: print('WARNING - masking not catching - must be either "masked" or "unmasked"')

            hdu.close()

        #flux_error_order = (flux_error_order)/(np.max(fluxes)-np.min(fluxes))
        #print('flux error: %s'%flux_error_order)
        #fluxes = (fluxes - np.min(fluxes))/(np.max(fluxes)-np.min(fluxes))
        #idx = tuple([fluxes>0])

        return np.array(fluxes), np.array(wavelengths), np.array(flux_error_order), sn ## for just LSD



#%%
import numpy as np
import math
import glob
import Mandel_Agol as ma
from scipy.interpolate import interp1d
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from astropy.io import fits
from matplotlib.pyplot import cm

#def combineccfs(spectra):
   # spectrum[:] = np.sum(spectra[:])/len(spectra[:])
    #return  spectrum


def gauss(x, rv, sd, height, cont):
    y = cont*(1-height*np.exp(-(x-rv)**2/(2*sd**2)))
    return y

def residualccfs(in_ccfs, in_ccfs_errors, master_out, master_out_errors, velocities):
    residuals=[]
    residual_errors = []
    #plt.figure('residuals')
    for i in range(len(in_ccfs)):
        ccf = in_ccfs[i]
        ccf_err = in_ccfs_errors[i]
        residual = (master_out+1)-(ccf+1)
        error = (np.sqrt(master_out_errors**2 + ccf_err**2))/np.sqrt(len(ccf_err))
        #residual = (ccf+1)/(master_out+1)-1
        #plt.scatter(velocities,residual)
        residuals.append(residual)
        residual_errors.append(error)
        '''
        plt.figure()
        plt.plot(ccf)
        plt.plot(master_out)
        #plt.plot(residual)
        plt.show()
        '''
        #plt.show()

    return residuals, residual_errors
####################################################################################################################################################################


##path = '/Users/lucydolan/Documents/CCF_method/HD189733_HARPS_CCFS/'
path = '/home/lsd/Documents/Starbase/novaprime/Documents/LSD_Figures/'
#path = '/Users/lucydolan/Starbase/LSD_Figures/'
month = 'August2007' #August, July, Sep
#path = '%s%s_master_out_LSD_profile.fits'%(save_path, month)

months = [#'August2007', 
          'July2007'#,
          #'July2006',
          #'Sep2006'
          ]

all_resi=[]
all_phase=[]
all_prof=[]
results_all = []

masters = []
for month in months:
    directory =  '%s'%(path)
    #file ='%s%s_master_out.fits'%(directory, month)
    file ='%s%s_master_out_LSD_profile.fits'%(directory, month)
    #ccf_file='%s%s_master_out_ccfs.fits'%(directory, month)
    #file ='%s%s_master_out_LSD_v3.fits'%(directory, month)
    #file ='%s%s_master_out_cegla.fits'%(directory, month)
    # all_ccfs = fits.open(ccf_file)
    all_profiles = fits.open(file)

    profile_spec = all_profiles[0].data[0]
    velocities = np.linspace(-10,10,len(profile_spec))


    count = 0
    rv_phases = []
    rv_results = []
    ccf_rv_results = []
    rvs = []
    fwhm = []
    plt.figure()
    st = 15
    end = -15
    st = 0
    end = len(velocities)
    # cmap = plt.colormaps('Blues')'

    # colour = cm.Blues(np.linspace(0, 1, len(all_profiles)))
    # plt.figure()
    # for i in range(len(all_profiles)-1):
    #     y = all_profiles[i].data[0]
    #     popt, pcov = curve_fit(gauss, velocities[st:end], y[st:end])
    #     perr= np.sqrt(np.diag(pcov))
    #     plt.plot(velocities[st:end], y[st:end], 'k')
    #     plt.plot(velocities[st:end], gauss(velocities[st:end], popt[0], popt[1], popt[2], popt[3]), 'r')
    #     rvs.append(popt[0])
    #     fwhm.append(2.355*popt[1])
    #     rv_phases.append(all_profiles[count].header['PHASE'])
        
    #     rv_results.append(all_profiles[count].header['RESULT'])
    #     count += 1
    # #plt.legend()

    # plt.figure('ACID and CCF RV (-median)', figsize = [9, 7])
    # plt.ylabel('RV - median(RV)')
    # plt.xlabel('Phase')
    # plt.scatter(rv_phases, rvs-np.median(rvs), label = 'ACID NEW', color = 'm')
    # # plt.scatter(rv_phases, rvs2-np.median(rvs2), label = 'ACID NEW_moremask', color = 'k')
    # plt.legend()
    # plt.show()

    # ccf_spec = all_ccfs[0].data[0]
    # ccf_velocities=all_ccfs[0].header['CRVAL1']+(np.arange(ccf_spec.shape[0]))*all_ccfs[0].header['CDELT1']
    # ccf_velocities = np.linspace(-19,19,len(ccf_spec))

    # master_position_ccf = len(all_ccfs)-1
    # master_out_ccf, master_out_errors_ccf= all_ccfs[master_position_ccf].data

    master_position = len(all_profiles)-1
    master_out, master_out_errors= all_profiles[master_position].data
    masters.append(master_out)

    plt.figure('master out')
    plt.plot(velocities, master_out, label = 'LSD')
    # plt.plot(ccf_velocities, master_out_ccf/master_out_ccf[0]-1, label = 'ccf')
    plt.legend()
    plt.show()

    # master_out_ccf = master_out_ccf/master_out_ccf[0]-1

    in_profiles = []
    in_profiles_errors = []
    phases = []
    results = []
    in_ccfs = []
    in_ccfs_errors = []
    phases_ccfs = []
    ccf_results = []
    out_profiles = []
    out_profile_error = []

    plt.figure('all_ccfs')
    # for line in range(0,master_position_ccf):
    #     ccf = all_ccfs[line].data[0]
    #     ccf_errors = all_ccfs[line].data[1]
    #     ccf_phase = all_ccfs[line].header['PHASE']
    #     in_ccfs.append(ccf/ccf[0]-1)
    #     in_ccfs_errors.append(ccf_errors/ccf[0])
    #     #plt.plot(ccf, label = '%s_%s'%(result, line))
    #     phases_ccfs.append(ccf_phase)
    #     #all_phase.append(phase)
    #     #ccf_results.append(ccf_result)
    for line in range(0, master_position):
        profile = all_profiles[line].data[0]
        profile_errors = all_profiles[line].data[1]
        phase = all_profiles[line].header['PHASE']
        result = all_profiles[line].header['RESULT']
        ##adding in M and A curve
        P=2.21857567 #Cegla et al, 2016 - days
        T=2454279.436714 #cegla et al,2016
        a_Rs = 8.786 #Cristo et al - 8.786
        b=0.687 #Cristo et al, 2022
        RpRs = 0.15667
        u1=0.816
        u2=0  #Sing et al, 2011
        i = 85.5*np.pi/180 #Cristo et al, 2022

        z = ma.MandelandAgol(phase, a_Rs, i)
        transit_curve = ma.occultquad(z, RpRs, [u1, u2])
        print(line)
        print(phase)
        print(transit_curve)
        print(result)

        # profile = (profile[5:-5]+1)*transit_curve
        # profile_errors = profile_errors[5:-5]*transit_curve
        # if line ==0:
        #     velocities = velocities[5:-5]
        # print(velocities.shape, profile.shape)

        profile = (profile+1)*transit_curve
        profile_errors = profile_errors*transit_curve
        
        in_profiles.append(profile)
        in_profiles_errors.append(profile_errors)
        plt.plot(velocities, profile, label = '%s_%s'%(result, line))
        phases.append(phase)
        all_phase.append(phase)
        if result == 'out':
            out_profiles.append(profile)
            out_profile_error.append(profile)
        results.append(result)
        results_all.append(result)
    plt.legend()
    plt.show()

    # plt.figure('ccfs - LSDs')
    # f2 = interp1d(ccf_velocities, in_ccfs, kind='linear', bounds_error=False, fill_value=np.nan)
    # plt.imshow(f2(velocities)-in_profiles)
    # plt.colorbar()
    # plt.show()

    #print(phases)
    # profile_spec = all_profiles[0].data[0]
    # #velocities=all_profiles[0].header['CRVAL1']+(np.arange(profile_spec.shape[0]))*all_profiles[0].header['CDELT1']
    # velocities = np.linspace(-15,15,len(profile_spec))

    # ccf_spec = all_ccfs[0].data[0]
    # ccf_velocities=all_ccfs[0].header['CRVAL1']+(np.arange(ccf_spec.shape[0]))*all_ccfs[0].header['CDELT1']


   # K = -2.277 #km/s - Boisse et al, 2009
    #velocities = velocities - K  ### Adjusting doppler reflex ###
    residual_profiles, residual_profile_errors = residualccfs(in_profiles, in_profiles_errors, master_out, master_out_errors, velocities)
    # plt.figure()
    # for out_prof in out_profiles:
    #     residual_profiles, residual_profile_errors = residualccfs(out_profiles, out_profile_error, out_prof, master_out_errors, velocities)
    #     # for resi_prof in residual_profiles:
    #     plt.plot(np.mean(residual_profiles, axis = 1))
    # plt.show()
    # residual_ccfs, residual_errors = residualccfs(in_ccfs, in_ccfs_errors, master_out_ccf, master_out_errors_ccf, ccf_velocities)

    '''
    plt.figure(month)
    outs=[]
    ins = []
    for ccf in residual_ccfs:
        if max(ccf)<0.4:
            plt.plot(ccf)
            ins.append(ccf)
        else:outs.append(ccf)
    print(len(outs))
    print(len(ins))
    '''
    # phases = np.array(phases)
    # results = np.array(results)
    # print(month)
    # plt.figure()
    # i=0
    # for ccf1 in residual_ccfs:
    #     for phase in phases:
    #         if phases_ccfs[i] == phase:
    #             print(results[tuple([phases==phase])])
    #             if results[tuple([phases==phase])]=='out':colour = '--'
    #             else:colour = '-'
    #             plt.plot(ccf_velocities, ccf1, label = '%s'%(i), linestyle = colour)
    #
    #     #plt.fill_between(ccf_velocities, ccf1-residual_errors[i], ccf1+residual_errors[i], alpha = 0.3)
    #     #all_resi.append(ccf1)
    #     i+=1
    # #plt.legend()
    # plt.show()

    print(month)
    plt.figure(month)
    i=0
    for ccf1 in residual_profiles:
        if results[i]=='out':colour = '--'
        else:colour = '-'
        plt.plot(velocities, ccf1, label = '%s_%s'%(results[i], i), linestyle = colour)
        plt.fill_between(velocities, ccf1-residual_profile_errors[i], ccf1+residual_profile_errors[i], alpha = 0.3)
        all_resi.append(ccf1+1)
        all_prof.append(in_profiles[i])
        i+=1
    plt.legend()
    plt.show()


    #write in data
    hdu=fits.HDUList()
    for data in residual_profiles:
        hdu.append(fits.PrimaryHDU(data=data))

    #hdu.append(fits.PrimaryHDU(data=master_out))

    plt.figure()
    #write in header
    for p in range(len(phases)):
        phase = phases[p]
        hdr=fits.Header()
        hdr['CRVAL1']=all_profiles[0].header['CRVAL1']
        hdr['CDELT1']=all_profiles[0].header['CDELT1']
        hdr['OBJECT']='HD189733b'
        hdr['NIGHT']='%s'%month
        hdr['K']=-2.277
        hdr['PHASE']=phase
        # hdu[p].header=hdr
        print(phase)
        print(results[p])
        
        if results[p] == 'out':
            plt.plot(residual_profiles[p], label = '%s, %s'%(phases[p], max(residual_profiles[p])-min(residual_profiles[p])))
    plt.legend()
    plt.show()
    #hdu.writeto('%s%s_residual_ccfs.fits'%(directory, month), output_verify='fix', overwrite = 'True')
    #hdu.writeto('%s%s_residual_ccfs_LSD_v3.fits'%(directory, month), output_verify='fix', overwrite = 'True')

n1 = len(all_resi)

col0=((np.round(np.linspace(0,170,num=n1))).astype(int))
#col1=((np.round(np.linspace(0,170,num=n2))).astype(int))
#col2=((np.round(np.linspace(50,220,num=n3))).astype(int))

#plt.cm.Reds(col1[i-n1])
#plt.cm.Greens(col2[i])
#plt.cm.Reds(col1[i])
#plt.cm.Blues(col0[i])
'''
fig, ax1 = plt.subplots(1,1)

for i in range(len(all_resi)):
    ax1.plot(velocities, all_resi[i],color=plt.cm.gist_rainbow(col0[i]),linewidth=0.7)
plt.show()
'''

fig, ax = plt.subplots(2)
#ax[0], ax[1] = fig.add_subplot()
ax[0].set_title('Residual LSD Profiles')
for ccf in all_resi:
    ax[0].plot(velocities, ccf)
ax[0].set_ylabel('Residual Flux')
#plt.savefig('/Users/lucydolan/Documents/CCF_method/Figures/residual_ccfs_mine')

cmap = plt.get_cmap('jet')
divider = make_axes_locatable(ax[1])
cax = divider.append_axes('right', size="7%", pad=0.2,)
im = ax[1].pcolormesh(velocities, all_phase, all_resi, cmap = cmap)
ax[1].set_ylim(-0.02, 0.02)
fig.colorbar(im,label='Residual Flux', cax = cax)
ax[1].set_xlabel('Velocity (km/s)')
ax[1].set_ylabel('Phase')
plt.show()

## Cegla style fitting of rvs
x = velocities#[1:]
all_resi = np.array(all_resi)
# all_resi = all_resi[:, 1:]
rvs = []
count = 0

P=2.21857567 #Cegla et al, 2006 - days
t=0.076125 #Torres et al, 2008 - days

rv_phases = []
for y in all_prof:
    #if (-t/(2*P)+0.001)<all_phase[count]<0.0155:
    
        popt, pcov = curve_fit(gauss, x, y)
        perr= np.sqrt(np.diag(pcov))
        rvs.append([popt[0], perr[0]])
        rv_phases.append(all_phase[count])
        count += 1

rvs = np.array(rvs)

plt.figure('RV measured from full profile')
plt.xlabel('Phase')
plt.ylabel('Local RV (km/s)')
plt.errorbar(rv_phases, rvs[:,0], yerr = rvs[:,1], fmt='o', label = 'LSD')

x = velocities#[1:]
all_resi = np.array(all_resi)
#all_resi = all_resi[:, 1:]
rvs = []
count = 0

P=2.21857567 #Cegla et al, 2006 - days
t=0.076125 #Torres et al, 2008 - days

rv_phases = []

plt.figure()
for y in all_resi:
    #if (-t/(2*P)+0.001)<all_phase[count]<0.0155:
    if np.max(y) - np.min(y)>0.006:
        # plt.figure()
        # plt.plot(x, y)
        # plt.show()
        try:
            popt, pcov = curve_fit(gauss, x, y)
            perr= np.sqrt(np.diag(pcov))
            rvs.append([popt[0], perr[0]])
            rv_phases.append(all_phase[count])
        except:
            print('could not fit phase: %s'%all_phase[count])
    else: 
        plt.plot(x, y)
        print(results_all[count])
    count += 1

rvs = np.array(rvs)

# plt.figure('RV measured from residual profile')
# plt.xlabel('Phase')
# plt.ylabel('Local RV (km/s)')
# plt.errorbar(rv_phases, rvs[:,0], yerr = rvs[:,1], fmt='o', label = 'LSD')
# plt.show()

plt.figure('master profiles')
for i in range(len(months)):
    plt.plot(velocities, masters[i], label = '%s'%months[i])
plt.legend()
plt.show()



# Old LSD alpha matrix code, iterative step:
# The below I think works similarly to above, but but uses only for loops, rather then using a for loop
# that only involves numpy calculations. With testing, the above is much faster.
else:
    warnings.warn('Large wavelength ranges give large computation time. Seperate wavelength range into smaller chunks for faster computation.', DeprecationWarning, stacklevel=2)
    alpha = np.zeros((len(blankwaves), len(velocities)))

    for j in tqdm(range(0, len(blankwaves)), desc='Calculating alpha matrix'):
        for i in (range(0, len(self.wavelengths_expected))):
            vdiff = ((blankwaves[j] - self.wavelengths_expected[i]) * ckms) / self.wavelengths_expected[i]
            if vdiff <= (np.max(velocities) + deltav) and vdiff >= (np.min(velocities) - deltav):
                diff = blankwaves[j] - self.wavelengths_expected[i]
                vel = const.c / 1e3 * (diff / self.wavelengths_expected[i])
                for k in range(0, len(velocities)):
                    x = (velocities[k] - vel) / deltav
                    if -1. < x and x < 0.:
                        delta_x = (1 + x)
                        alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
                    elif 0. <= x and x < 1.:
                        delta_x = (1 - x)
                        alpha[j, k] = alpha[j, k] + depths_expected[i] * delta_x
            else:
                pass

# Vectorized old part:
# The below was simplified using np.clip (as shown above)
            alpha_mask_1 = np.logical_and(-1. < x, x < 0.)
            alpha_mask_2 = np.logical_and(0. <= x, x < 1.)
            delta_x_1 = 1 + x
            delta_x_2 = 1 - x
            delta_x_1[alpha_mask_1==False] = 0
            delta_x_2[alpha_mask_2==False] = 0
            # Update alpha array using calculated delta_x values
            alpha = np.zeros((len(blankwaves), len(velocities)))
            alpha += (depths_expected[:, np.newaxis] * delta_x_1).sum(axis=1)
            alpha += (depths_expected[:, np.newaxis] * delta_x_2).sum(axis=1)

# Old blaze correct code part 1:
# ## test - s1d interpolated onto e2ds wavelength grid ##
                # hdu_e2ds=fits.open('%s'%file.replace('s1d', 'e2ds'))
                # spec_e2ds=hdu_e2ds[0].data
                # header_e2ds=hdu_e2ds[0].header

                # wave_e2ds=get_wave(spec_e2ds, header_e2ds)*(1.+brv/2.99792458e5)

                # # plt.figure()
                # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), wave_e2ds[order][1:]-wave_e2ds[order][:-1], label = 'e2ds wave (after berv)')
                # # plt.scatter(np.arange(len(wave_e2ds[order][:-1])), get_wave(spec_e2ds, header_e2ds)[order][1:]-get_wave(spec_e2ds, header_e2ds)[order][:-1], label = 'e2ds wave (before berv)')
                # # # plt.scatter(np.arange(len(wavelengths[:-1])), wavelengths[:-1]-wavelengths[1:], label = 's1d wave')
                # # plt.legend()
                # # plt.show()

                # # id = np.logical_and(wave_e2ds<np.max(wavelengths), wave_e2ds>np.min(wavelengths))
                # # print(wave_e2ds*u.AA)
                # # print(wavelengths*u.AA)
                # # print(fluxes*u.photon)

                # blaze_file = glob.glob('%sblaze_folder/**blaze_A*.fits'%(directory))
                # # print('%sblaze_folder/**blaze_A*.fits'%(directory))
                # # print(blaze_file)
                # blaze_file = blaze_file[0]
                # blaze =fits.open('%s'%blaze_file)
                # blaze_func = blaze[0].data
                # spec_e2ds = spec_e2ds/blaze_func
            
                # diff_arr = wavelengths[1:] - wavelengths[:-1]
                # print(diff_arr)
                # wavelengths = wavelengths[:-1]
                # fluxes = fluxes[:-1]/diff_arr

                # s1d_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
                # fluxcon = FluxConservingResampler()
                # new_spec = fluxcon(s1d_spec, wave_e2ds[order]*u.AA)

                # wavelengths = new_spec.spectral_axis
                # fluxes = new_spec.flux

                # wavelengths = wavelengths[10:len(wave_e2ds[order])-9]/u.AA
                # fluxes = fluxes[10:len(wave_e2ds[order])-9]/u.Unit('photon AA-1')
                # flux_error_order = flux_error_order[10:len(wave_e2ds[order])-10]

                # diff_arr = wavelengths[1:] - wavelengths[:-1]
                # print(diff_arr)
                # wavelengths = wavelengths[:-1]
                # fluxes = fluxes[:-1]*diff_arr

                # print(wavelengths)
                # print(fluxes)

                # plt.figure()
                # plt.title('interpolated s1d comapred to actual e2ds spectrum')
                # plt.plot(wavelengths, fluxes, label = 'interpolated s1d on e2ds wave grid')
                # plt.plot(wave_e2ds[order], spec_e2ds[order], label = 'e2ds spectrum')
                # plt.legend()
                # plt.show()


                # ## end of test ##

                # ## test2 - synthetic s1d spectrum - using wavelength grid ##
                # def gauss(x1, rv, sd, height, cont):
                #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
                #     return y1

                # wavelength_grid = wavelengths
                # flux_grid = np.ones(wavelength_grid.shape)

                # linelist = '/home/lsd/Documents/fulllinelist0001.txt'
                
                # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
                # wavelengths_expected1 =np.array(linelist_expected[:,0])
                # depths_expected1 = np.array(linelist_expected[:,1])
                # # print(len(depths_expected1))

                # wavelength_min = np.min(wavelengths)
                # wavelength_max = np.max(wavelengths)

                # print(wavelength_min, wavelength_max)

                # wavelengths_expected=[]
                # depths_expected=[]
                # no_line =[]
                # for some in range(0, len(wavelengths_expected1)):
                #     line_min = 0.25
                #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
                #         wavelengths_expected.append(wavelengths_expected1[some])
                #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
                #         depths_expected.append(depths_expected1[some])
                #     else:
                #         pass
                
                # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
                # count_range = np.array(count_range, dtype = int)
                # print(count_range)
                # vgrid = np.linspace(-21,18,48)
                # try: ccf = fits.open(file.replace('s1d', 'ccf_K5'))
                # except: ccf = fits.open(file.replace('s1d', 'ccf_G2'))
                # rv = ccf[0].header['ESO DRS CCF RV']

                
                # for line in count_range:
                #     mid_wave = wavelengths_expected[line]
                #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
                #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
                #     prof_wavelength_grid = wavelength_grid[id]
                #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
                #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
                #     # plt.figure()
                #     # plt.plot(prof_wavelength_grid, prof)
                #     # plt.show()
                #     flux_grid[id] = prof

                # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
                # poly = np.poly1d(coeffs*fluxes[0])
                # fit = poly(wavelengths)

                # wavelengths = wavelength_grid
                # fluxes = flux_grid * fit

# Old blaze correct code part 2:

# plt.figure()
            # plt.figure('blaze for orders 28, 29 and 30')
            # plt.plot(wave[28], blaze[0].data[28])
            # plt.plot(wave[29], blaze[0].data[29])
            # plt.plot(wave[30], blaze[0].data[30])

            # plt.figure()
            # plt.title('blaze for orders 28, 29 and 30 summed together')
            # pixel_grid = np.linspace(np.min(wave[28]), np.max(wave[30]), len(np.unique(wave[28:30])))
            # blaze_sum = np.zeros(pixel_grid.shape)

            # f28 = interp1d(wave[28], blaze[0].data[28], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
            # f29 = interp1d(wave[29], blaze[0].data[29], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')
            # f30 = interp1d(wave[30], blaze[0].data[30], kind = 'linear', bounds_error=False, fill_value = 'extrapolate')

            # idx28 = np.logical_and(pixel_grid>np.min(wave[28]), pixel_grid<np.max(wave[28]))
            # idx29 = np.logical_and(pixel_grid>np.min(wave[29]), pixel_grid<np.max(wave[29]))
            # idx30 = np.logical_and(pixel_grid>np.min(wave[30]), pixel_grid<np.max(wave[30]))

            # blaze_28 = f28(pixel_grid[idx28])
            # blaze_29 = f29(pixel_grid[idx29])
            # blaze_30 = f30(pixel_grid[idx30])

            # wave28 = pixel_grid[idx28]
            # wave29 = pixel_grid[idx29]
            # wave30 = pixel_grid[idx30]

            # for pixel in range(len(pixel_grid)):
            #     wavell = pixel_grid[pixel]
                
            #     idx28 = tuple([wave28==wavell])
            #     idx29 = tuple([wave29==wavell])
            #     idx30 = tuple([wave30==wavell])
                
            #     try: b = blaze_28[idx28][0]
            #     except: b=0
            #     try: b1 = blaze_29[idx29][0]
            #     except: b1=0
            #     try: b2 = blaze_30[idx30][0]
            #     except: b2=0

            #     print(wavell, b, b1, b2)

            #     blaze_sum[pixel] = b + b1 + b2 

            # plt.plot(pixel_grid, blaze_sum)
            # plt.show()

            # plt.figure()
            # plt.title('e2ds after blaze orders 28, 29 and 30')
            # plt.plot(wave[28], spec[28])
            # plt.plot(wave[29], spec[29])
            # plt.plot(wave[30], spec[30])
            
            ## TEST - adjusting e2ds spectrum onto s1d continuum ##

            # ## first interpolate s1d onto e2ds wavelength grid ##
            # s1d_file = fits.open(file.replace('e2ds', 's1d'))
            # s1d_spec = s1d_file[0].data
            # wave_s1d = s1d_file[0].header['CRVAL1']+(np.arange(s1d_spec.shape[0]))*s1d_file[0].header['CDELT1']

            # wavelengths = wave_s1d
            # fluxes = s1d_spec

            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 'e2ds spectrum - corrected to s1d continuum')
            # plt.plot(wave_s1d, s1d_spec, label = 's1d spectrum')
            # plt.legend()
            # plt.show()

            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 's1d')
            # plt.legend()

            # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # print(diff_arr)
            # wavelengths = wavelengths[:-1]
            # fluxes = fluxes[:-1]

            # fluxes = fluxes/diff_arr

            # fluxes = np.ones(fluxes.shape)
            # for i in range(len(fluxes)):
            #     fluxes[i] = fluxes[i]*8000
            # plt.figure()
            # plt.plot(wavelengths, fluxes, label = 's1d in photons AA-1')
            # #plt.legend()

            # interpolate s1d onto e2ds wavlengths - non flux conserving
            # s1dd_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
            # fluxcon = FluxConservingResampler()
            # extended_e2ds_wave = np.concatenate((wave[order], [wave[order][-1]+0.01]))

            ## MM-LSD way

            # fluxes = spec[order]
            # flux_error_order = flux_error[order]
            # wavelengths = wave[order]

            # for i in range(len(fluxes)):
            #     if i ==0:
            #         # print(fluxes[i])
            #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[1]-wavelengths[0]))
            #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
            #         # print(fluxes[i])
            #     else:
            #         # print(fluxes[i])
            #         fluxes[i] = fluxes[i]*(0.01/(wavelengths[i]-wavelengths[i-1]))
            #         # print(0.01/(2.99792458e5*(wavelengths[1]-wavelengths[0])))
            #         # print(fluxes[i])

            # ## end of MM-LSD way

            # new_spec = fluxcon(s1dd_spec, wavelengths*u.AA)

            # reference_wave = new_spec.spectral_axis/u.AA
            # reference_flux = new_spec.flux/u.Unit('photon AA-1')

            # # print(len(wavelengths_new))
            # # print(len(wave[order]))

            # # # plt.plot(wavelengths_new, fluxes, label = 'interpolated s1d in photons AA-1')
            # # # plt.legend()

            # #diff_arr = wavelengths_new[1:] - wavelengths_new[:-1]
            # #reference_wave = wavelengths_new[:-1]
            # #reference_flux = fluxes[:-1]*diff_arr
            
            # # # print(len(reference_wave))
            # # # print(len(wave[order]))

            # # # print(reference_wave-wave[order])

            # # # plt.figure()
            # # # plt.plot(reference_wave, reference_flux, label = 'interpolated s1d in photons per bin')
            # # # plt.legend()
            # # # plt.show()
            # # ## divide e2ds spectrum by interpolated s1d and fit polynomial to result

            # reference_wave = np.array(reference_wave, dtype = float)
            # reference_flux = np.array(reference_flux, dtype = float)
            # div_frame = fluxes/reference_flux

            # # plt.figure()
            # # plt.plot(reference_wave, div_frame)
            # # plt.show()

            # # # ### creating windows to fit polynomial to
            # # # binned = np.zeros(int(len(div_frame)/2))
            # # # binned_waves = np.zeros(int(len(div_frame)/2))
            # # # for i in range(0, len(div_frame)-1, 2):
            # # #     pos = int(i/2)
            # # #     binned[pos] = (div_frame[i]+div_frame[i+1])/2
            # # #     binned_waves[pos] = (reference_wave[i]+reference_wave[i+1])/2

            # # # plt.plot(frame_wavelengths[n], frames_unadjusted[n], color = 'b', label = 'unadjusted')
            # # # plt.figure()
            # # # plt.plot(frame_wavelengths[n], frames[n])
            # # # plt.show()

            # ### fitting polynomial to div_frame
            # coeffs=np.polyfit(reference_wave, div_frame, 3)
            # poly = np.poly1d(coeffs)
            # # print(coeffs)
            # inputs = coeffs[::-1]
            # # print(inputs)

            # wavelengths = reference_wave

            # fit = poly(wavelengths)
        
            # # # plt.figure()
            # # # plt.plot(reference_wave, reference_flux, label= 'reference')
            # # # plt.plot(wave[order], spec[order], label = 'e2ds')
            # # # plt.legend()

            # # # plt.figure()
            # # # plt.plot(reference_wave-wavelengths[:-1], label = 'reference_wave-wavelengths')
            # # # plt.legend()
            # # # plt.show()

            # # plt.figure()
            # # plt.scatter(reference_wave, div_frame, label = 'div flux')
            # # plt.plot(wavelengths, fit, label = 'fit')
            # # # plt.plot(reference_wave, poly(reference_wave), label = 'poly(reference_wave)')
            # # plt.legend()
            # # plt.show()
            
            # fluxes = spec[order]/fit
            # flux_error_order = flux_error[order]/fit

            # # plt.figure()
            # # plt.plot(wavelengths, fluxes/reference_flux, label = 'continuum adjusted e2ds/s1d')
            # # plt.legend()
            # # plt.show()

            # # plt.figure()
            # # plt.plot(wavelengths, fluxes-reference_flux, label = 'continuum adjusted e2ds-s1d')
            # # plt.legend()
            # # plt.show()
            # # plt.figure()
            # # plt.plot(wavelengths, spec[order], label = 'before')
            # # plt.plot(wavelengths, fit)
            # # plt.plot(wavelengths, fluxes, label= 'after')
            # # plt.legend()
            # # plt.show()

            # # idx_full = np.logical_and(wave_s1d>np.min(wave[28]), wave_s1d<np.max(wave[30]))
            # # plt.plot(wave_s1d[idx_full], s1d_spec[idx_full])
            # # plt.show()

            # blaze.close()
            # # plt.figure('after blaze correction - e2ds vs s1d - after berv correction')
            # # plt.plot(wave[order], spec[order], label = 'e2ds')
            # # plt.plot(wave_s1d[wave_s1d>np.max(wave[order])], spec_s1d[wave_s1d>np.max(wave[order])], label = 's1d')
            # # plt.show()

            

            # # # test - e2ds interpolated onto s1d wavelength grid ##
            # # hdu_s1d=fits.open('%s'%file.replace('e2ds', 's1d'))
            # # spec_s1d=hdu_s1d[0].data
            # # header_s1d=hdu_s1d[0].header

            # # wave_s1d=header_s1d['CRVAL1']+(header_s1d['CRPIX1']+np.arange(spec_s1d.shape[0]))*header_s1d['CDELT1']
            # # id = np.logical_and(wave_s1d<np.max(wavelengths), wave_s1d>np.min(wavelengths))
            # # print(wave_s1d*u.AA)
            # # print(wavelengths*u.AA)
            # # print(fluxes*u.Unit('erg cm-2 s-1 AA-1'))
            # # # plt.figure('s1d compared to interpolated e2ds')
            # # # plt.title('s1d compared to interpolated e2ds')
            # # # plt.plot(wave_s1d, spec_s1d, label = 's1d spectrum')

            # # ## these fluxes are in photons per bin - I need them in photons per Angstrom
            # # ## therefore i do flux/angstroms in pixel
            # # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # print(diff_arr)
            # # wavelengths = wavelengths[:-1]
            # # fluxes = fluxes[:-1]
            # # # plt.figure('changing flux units')
            # # # plt.plot(wavelengths, fluxes, label = 'flux per pixel')

            # # fluxes = fluxes/diff_arr

            # # # plt.plot(wavelengths, fluxes, label = ' flux per A')

            # # e2ds_spec = Spectrum1D(spectral_axis = wavelengths*u.AA, flux = fluxes*u.Unit('photon AA-1'))
            # # fluxcon = FluxConservingResampler()
            # # new_spec = fluxcon(e2ds_spec, wave_s1d[id]*u.AA)

            # # wavelengths = new_spec.spectral_axis
            # # fluxes = new_spec.flux

            # # wavelengths = wavelengths[:4097]/u.AA
            # # fluxes = fluxes[:4097]/u.Unit('photon AA-1')

            # # diff_arr = wavelengths[1:] - wavelengths[:-1]
            # # wavelengths = wavelengths[:-1]
            # # fluxes = fluxes[:-1]*diff_arr

            # # # plt.figure('s1d compared to interpolated e2ds') 
            # # # plt.plot(wavelengths, fluxes, label = 'interpolated e2ds onto s1d')
            # # # plt.xlim(np.min(wavelengths), np.max(wavelengths))
            # # # plt.legend()
            # # # plt.show() 

            # # print(wavelengths)
            # # print(fluxes)

            # #end of test ##

            # # # test2 - synthetic e2ds spectrum - using wavelength grid ##
            # # def gauss(x1, rv, sd, height, cont):
            # #     y1 = height*np.exp(-(x1-rv)**2/(2*sd**2)) + cont
            # #     return y1

            # # wavelength_grid = wavelengths
            # # flux_grid = np.ones(wavelength_grid.shape)

            # # linelist = '/home/lsd/Documents/fulllinelist0001.txt'

            # # linelist_expected = np.genfromtxt('%s'%linelist, skip_header=4, delimiter=',', usecols=(1,9))
            # # wavelengths_expected1 =np.array(linelist_expected[:,0])
            # # depths_expected1 = np.array(linelist_expected[:,1])
            # # # print(len(depths_expected1))

            # # wavelength_min = np.min(wavelengths)
            # # wavelength_max = np.max(wavelengths)

            # # print(wavelength_min, wavelength_max)

            # # wavelengths_expected=[]
            # # depths_expected=[]
            # # no_line =[]
            # # for some in range(0, len(wavelengths_expected1)):
            # #     line_min = 0.25
            # #     if wavelengths_expected1[some]>=wavelength_min and wavelengths_expected1[some]<=wavelength_max and depths_expected1[some]>=line_min:
            # #         wavelengths_expected.append(wavelengths_expected1[some])
            # #         #depths_expected.append(depths_expected1[some]+random.uniform(-0.1, 0.1))
            # #         depths_expected.append(depths_expected1[some])
            # #     else:
            # #         pass

            # # count_range = np.array([len(wavelengths_expected)]*10)*np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])
            # # count_range = np.array(count_range, dtype = int)
            # # print(count_range)
            # # vgrid = np.linspace(-21,18,48)
            # # try: ccf = fits.open(file.replace('e2ds', 'ccf_K5'))
            # # except: ccf = fits.open(file.replace('e2ds', 'ccf_G2'))
            # # rv = ccf[0].header['ESO DRS CCF RV']


            # # for line in count_range:
            # #     mid_wave = wavelengths_expected[line]
            # #     wgrid = 2.99792458e5*mid_wave/(2.99792458e5-vgrid)
            # #     id = np.logical_and(wavelength_grid<np.max(wgrid), wavelength_grid>np.min(wgrid))
            # #     prof_wavelength_grid = wavelength_grid[id]
            # #     prof_v_grid = ((prof_wavelength_grid - mid_wave)*2.99792458e5)/prof_wavelength_grid
            # #     prof = gauss(prof_v_grid, rv, 2.47, -depths_expected[line], 1.)
            # #     # id = tuple([prof_v_grid<-0.99])
            # #     # plt.figure()
            # #     # plt.plot(prof_wavelength_grid, prof)
            # #     # plt.show()
            # #     flux_grid[id] = prof

            # # # plt.figure()
            # # # plt.plot(wavelength_grid, flux_grid)
            # # # plt.show()

            # # coeffs=np.polyfit(wavelengths, fluxes/fluxes[0], 3)
            # # poly = np.poly1d(coeffs*fluxes[0])
            # # fit = poly(wavelengths)

            # # wavelengths = wavelength_grid
            # # fluxes = flux_grid * fit

            # # plt.figure()
            # plt.plot(wavelengths, fluxes)
            # plt.show()

            # find overlapping regions 
            # last_wavelengths = wave[order-1]
            # next_wavelengths = wave[order+1]
            # last_spec = spec[order-1]
            # next_spec = spec[order+1]
            # last_error = flux_error[order-1]
            # next_error = flux_error[order+1]
            # min_overlap = np.min(wavelengths)
            # max_overlap = np.max(wavelengths)

            
            # idx_ = tuple([wavelengths>min_overlap])
            # last_idx = np.logical_and(last_wavelengths>min_overlap, last_wavelengths<max_overlap)
            # next_idx = np.logical_and(next_wavelengths>min_overlap, next_wavelengths<max_overlap)
            
            # overlap = np.array(([list(last_wavelengths[last_idx]), list(last_spec[last_idx]), list(last_error[last_idx])], [list(next_wavelengths[next_idx]), list(next_spec[next_idx]), list(next_error[next_idx])]))
            
            # # overlap[0, 0] = list(last_wavelengths[last_idx])
            # # overlap[0, 1] = list(last_spec[last_idx])
            # # overlap[1, 0] = list(next_wavelengths[next_idx])
            # # overlap[1, 1] = list(next_spec[next_idx])

            # print(overlap)
            # plt.figure()
            # plt.plot(wavelengths, fluxes)
            # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
            # plt.show()
            # wavelengths = wavelengths[idx]
            # fluxes = fluxes[idx]
            # flux_error_order = flux_error_order[idx]

# Telluric correction:
## telluric correction
        # tapas = fits.open('/Users/lucydolan/Starbase/tapas_000001.fits')
        # tapas_wvl = (tapas[1].data["wavelength"]) * 10.0
        # tapas_trans = tapas[1].data["transmittance"]
        # tapas.close()
        # brv=header['ESO DRS BERV']
        # tapas_wvl = tapas_wvl[::-1]/(1.+brv/2.99792458e5)
        # tapas_trans = tapas_trans[::-1]

        # background = upper_envelope(tapas_wvl, tapas_trans)
        # f = interp1d(tapas_wvl, tapas_trans / background, bounds_error=False)

        # plt.figure('telluric spec and real spec')
        # plt.plot(wavelengths, continuumfit(wavelengths, fluxes, 3))
        # plt.plot(wavelengths, f(wavelengths))
        # plt.show()
        
        # plt.figure()
        # plt.plot(tapas_wvl, tapas_trans)
        # plt.show()
        # print('overlap accounted for')

# Overlap stuff?
# print(len(wavelengths))
                # print(np.max(wavelengths), np.min(wavelengths))
                # print(min_overlap)
                # print(max_overlap)
                # idx_overlap = np.logical_and(wavelengths>=min_overlap, wavelengths<=max_overlap)
                # idx_overlap = tuple([idx_overlap==False])
                # overlap = []

                # plt.figure()
                # plt.plot(wavelengths, fluxes, label = 'Flux')
                # # plt.plot(wavelengths[idx_overlap], fluxes[idx_overlap])
                # plt.show()

                # plt.plot(wavelengths, fluxes, label = 's1d')
                # plt.legend()
                # plt.show()